# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import struct
from spinn_utilities.log import FormatAdapter
from spinn_machine import CoreSubsets
from spinnman.model.enums import CPUState
from data_specification import utility_calls
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from .constants import MICRO_TO_MILLISECOND_CONVERSION

logger = FormatAdapter(logging.getLogger(__name__))
_n_word_structs = []


def locate_extra_monitor_mc_receiver(
        machine, placement_x, placement_y,
        packet_gather_cores_to_ethernet_connection_map):
    """ Get the data speed up gatherer that can be used to talk to a\
        particular chip. This will be on the same board.

    :param ~spinn_machine.Machine machine: The machine descriptor
    :param int placement_x: The X coordinate of the reference chip
    :param int placement_y: The Y coordinate of the reference chip
    :param packet_gather_cores_to_ethernet_connection_map:
    :type packet_gather_cores_to_ethernet_connection_map:
        dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
    :rtype: DataSpeedUpPacketGatherMachineVertex
    """
    chip = machine.get_chip_at(placement_x, placement_y)
    return packet_gather_cores_to_ethernet_connection_map[
        chip.nearest_ethernet_x, chip.nearest_ethernet_y]


def read_data(x, y, address, length, data_format, transceiver):
    """ Reads and converts a single data item from memory.

    :param int x: chip x
    :param int y: chip y
    :param int address: base address of the SDRAM chip to read
    :param int length: length to read
    :param str data_format:
        the format to read memory (see :py:func:`struct.pack`)
    :param ~spinnman.transceiver.Transceiver transceiver:
        the SpinnMan interface
    """
    # pylint: disable=too-many-arguments

    data = transceiver.read_memory(x, y, address, length)
    return struct.unpack_from(data_format, data)[0]


def write_address_to_user0(txrx, x, y, p, address):
    """ Writes the given address into the user_0 register of the given core.

    :param ~spinnman.transceiver.Transceiver txrx: The transceiver.
    :param int x: Chip coordinate.
    :param int y: Chip coordinate.
    :param int p: Core ID on chip.
    :param int address: Value to write (32-bit integer)
    """
    user_0_address = txrx.get_user_0_register_address_from_core(p)
    txrx.write_memory(x, y, user_0_address, address)


def locate_memory_region_for_placement(placement, region, transceiver):
    """ Get the address of a region for a placement.

    :param int region: the region to locate the base address of
    :param ~pacman.model.placements.Placement placement:
        the placement object to get the region address of
    :param ~spinnman.transceiver.Transceiver transceiver:
        the python interface to the SpiNNaker machine
    :return: the address
    :rtype: int
    """
    regions_base_address = transceiver.get_cpu_information_from_core(
        placement.x, placement.y, placement.p).user[0]

    # Get the position of the region in the pointer table
    region_offset = utility_calls.get_region_base_address_offset(
        regions_base_address, region)

    # Get the actual address of the region
    return transceiver.read_word(placement.x, placement.y, region_offset)


def convert_string_into_chip_and_core_subset(cores):
    """ Translate a string list of cores into a core subset

    :param cores:
        string representing down cores formatted as x,y,p[:x,y,p]*
    :type cores: str or None
    :rtype: ~spinn_machine.CoreSubsets
    """
    ignored_cores = CoreSubsets()
    if cores is not None and cores != "None":
        for downed_core in cores.split(":"):
            x, y, processor_id = downed_core.split(",")
            ignored_cores.add_processor(int(x), int(y), int(processor_id))
    return ignored_cores


def flood_fill_binary_to_spinnaker(executable_targets, binary, txrx, app_id):
    """ Flood fills a binary to spinnaker on a given `app_id` \
        given the executable targets and binary.

    :param ~spinnman.model.ExecutableTargets executable_targets:
        the executable targets object
    :param str binary: the (name of the) binary to flood fill
    :param ~spinnman.transceiver.Transceiver txrx: spinnman instance
    :param int app_id: the application ID to load it as
    :return: the number of cores it was loaded onto
    :rtype: int
    """
    core_subset = executable_targets.get_cores_for_binary(binary)
    txrx.execute_flood(
        core_subset, binary, app_id, wait=True, is_filename=True)
    return len(core_subset)


def read_config(config, section, item):
    """ Get the string value of a config item, returning None if the value\
        is "None"

    :param ~configparser.ConfigParser config:
        The configuration to look things up in.
    :param str section: The section name
    :param str item: The item name.
    :rtype: str or None
    """
    value = config.get(section, item)
    if value == "None":
        return None
    return value


def read_config_int(config, section, item):
    """ Get the integer value of a config item, returning None if the value\
        is "None"

    :param ~configparser.ConfigParser config:
        The configuration to look things up in.
    :param str section: The section name
    :param str item: The item name.
    :rtype: int or None
    """
    value = read_config(config, section, item)
    if value is None:
        return value
    return int(value)


def read_config_float(config, section, item):
    """ Get the float value of a config item, returning None if the value\
        is "None"

    :param ~configparser.ConfigParser config:
        The configuration to look things up in.
    :param str section: The section name
    :param str item: The item name.
    :rtype: float or None
    """
    value = read_config(config, section, item)
    if value is None:
        return value
    return float(value)


_BOOLEAN_STATES = {
    'true': True, '1': True, 'on': True, 'yes': True,
    'false': False, '0': False, 'off': False, 'no': False}


def read_config_boolean(config, section, item):
    """ Get the boolean value of a config item, returning None if the value\
        is "None"

    :param ~configparser.ConfigParser config:
        The configuration to look things up in.
    :param str section: The section name
    :param str item: The item name.
    :rtype: bool or None
    """
    value = read_config(config, section, item)
    if value is None:
        return value
    if value.lower() in _BOOLEAN_STATES:
        return _BOOLEAN_STATES[value.lower()]
    raise ValueError("Unknown boolean value {} in configuration {}:{}".format(
        value, section, item))


def generate_unique_folder_name(folder, filename, extension):
    """ Generate a unique file name with a given extension in a given folder

    :param str folder: where to put this unique file
    :param str filename:
        the name of the first part of the file without extension
    :param str extension: extension of the file
    :return: file path with a unique addition
    :rtype: str
    """
    new_file_path = os.path.join(folder, "{}{}".format(filename, extension))
    count = 2
    while os.path.exists(new_file_path):
        new_file_path = os.path.join(
            folder, "{}_{}{}".format(filename, count, extension))
        count += 1
    return new_file_path


def get_ethernet_chip(machine, board_address):
    """ Locate the chip with the given board IP address

    :param ~spinn_machine.Machine machine: the SpiNNaker machine
    :param str board_address: the board address to locate the chip of.
    :return: The chip that supports that board address
    :rtype: ~spinn_machine.Chip
    :raises ConfigurationException:
        when that board address has no chip associated with it
    """
    for chip in machine.ethernet_connected_chips:
        if chip.ip_address == board_address:
            return chip
    raise ConfigurationException(
        "cannot find the Ethernet connected chip with the board address {}"
        .format(board_address))


def convert_time_diff_to_total_milliseconds(sample):
    """ Convert between a time diff and total milliseconds.

    :param ~datetime.timedelta sample:
    :return: total milliseconds
    :rtype: float
    """
    return ((sample.total_seconds() * MICRO_TO_MILLISECOND_CONVERSION) +
            (sample.microseconds / MICRO_TO_MILLISECOND_CONVERSION))


def determine_flow_states(executable_types, no_sync_changes):
    """ Get the start and end states for these executable types.

    :param dict(ExecutableType,any) executable_types:
        the execute types to locate start and end states from
    :param int no_sync_changes: the number of times sync signals been sent
    :return: dict of executable type to states.
    :rtype: tuple(dict(ExecutableType,~spinnman.model.enums.CPUState),
        dict(ExecutableType,~spinnman.model.enums.CPUState))
    """
    expected_start_states = dict()
    expected_end_states = dict()
    for start_type in executable_types.keys():

        # cores that ignore all control and are just running
        if start_type == ExecutableType.RUNNING:
            expected_start_states[ExecutableType.RUNNING] = [
                CPUState.RUNNING, CPUState.FINISHED]
            expected_end_states[ExecutableType.RUNNING] = [
                CPUState.RUNNING, CPUState.FINISHED]

        # cores that require a sync barrier
        elif start_type == ExecutableType.SYNC:
            expected_start_states[ExecutableType.SYNC] = [CPUState.SYNC0]
            expected_end_states[ExecutableType.SYNC] = [CPUState.FINISHED]

        # cores that use our sim interface
        elif start_type == ExecutableType.USES_SIMULATION_INTERFACE:
            if no_sync_changes % 2 == 0:
                expected_start_states[start_type] = [CPUState.SYNC0]
            else:
                expected_start_states[start_type] = [CPUState.SYNC1]
            expected_end_states[start_type] = [CPUState.PAUSED]

    # if no states, go boom.
    if not expected_start_states:
        raise ConfigurationException(
            "Unknown executable start types {}".format(executable_types))
    return expected_start_states, expected_end_states


def convert_vertices_to_core_subset(vertices, placements):
    """ Converts vertices into core subsets.

    :param iterable(~pacman.model.graphs.machine.MachineVertex) vertices:
        the vertices to convert to core subsets
    :param ~pacman.model.placements.Placements placements:
        the placements object
    :return: the CoreSubSets of the vertices
    :rtype: ~spinn_machine.CoreSubsets
    """
    core_subsets = CoreSubsets()
    for vertex in vertices:
        placement = placements.get_placement_of_vertex(vertex)
        core_subsets.add_processor(placement.x, placement.y, placement.p)
    return core_subsets


def n_word_struct(n_words):
    """ Manages a precompiled cache of structs for parsing blocks of words. \
        Thus, this::

        data = n_word_struct(n_words).unpack(data_blob)

    Is much like doing this::

        data = struct.unpack("<{}I".format(n_words), data_blob)

    except quite a bit more efficient because things are shared including the
    cost of parsing the format.

    :param int n_words: The number of *SpiNNaker words* to be handled.
    :return: A struct for working with that many words.
    :rtype: ~struct.Struct
    """
    global _n_word_structs
    while len(_n_word_structs) < n_words + 1:
        _n_word_structs += [None] * (n_words + 1 - len(_n_word_structs))
    s = _n_word_structs[n_words]
    if s is not None:
        return s
    new_struct = struct.Struct("<{}I".format(n_words))
    _n_word_structs[n_words] = new_struct
    return new_struct


def get_defaultable_source_id(entry):
    """ Hack to support the source requirement for the router compressor\
        on chip.

    :param ~spinn_machine.MulticastRoutingEntry entry:
        the multicast router table entry.
    :return: return the source value
    :rtype: int
    """
    if entry.defaultable:
        return (list(entry.link_ids)[0] + 3) % 6
    elif entry.link_ids:
        return list(entry.link_ids)[0]
    return 0



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities.globals_variables import get_simulator


class FinaliseTimingData(object):
    """ Produces the timing information for the run.
    """

    __slots__ = []

    def __call__(self):
        """
        :return:
            mapping_time, dsg_time, load_time, execute_time, extraction_time
        :rtype: tuple(float, float, float, float, float)
        """
        # Note that this algorithm "knows" the simulator is an
        # AbstractSpinnakerBase instance of some kind, and that that the
        # _end_of_run_timing method exists.
        return get_simulator()._end_of_run_timing()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar


class LoadFixedRoutes(object):
    """ Load a set of fixed routes onto a SpiNNaker machine.
    """

    def __call__(self, fixed_routes, transceiver, app_id):
        """
        :param fixed_routes:
        :type fixed_routes:
            dict(tuple(int,int),~spinn_machine.FixedRouteEntry)
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param int app_id:
        """
        progress_bar = ProgressBar(
            total_number_of_things_to_do=len(fixed_routes),
            string_describing_what_being_progressed="loading fixed routes")

        for chip_x, chip_y in progress_bar.over(fixed_routes.keys()):
            transceiver.load_fixed_route(
                chip_x, chip_y, fixed_routes[chip_x, chip_y], app_id)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from threading import Thread
from spinn_utilities.log import FormatAdapter
from spinnman.exceptions import (
    SpinnmanIOException, SpinnmanInvalidPacketException,
    SpinnmanTimeoutException)
from spinnman.messages.eieio.command_messages import EIEIOCommandHeader
from spinnman.connections.udp_packet_connections import UDPConnection
from spinnman.constants import EIEIO_COMMAND_IDS as CMDS
from spinn_front_end_common.utilities.constants import NOTIFY_PORT
from .database_reader import DatabaseReader

logger = FormatAdapter(logging.getLogger(__name__))


class DatabaseConnection(UDPConnection):
    """ A connection from the toolchain which will be notified when the \
        database has been written, and can then respond when the database \
        has been read, and further wait for notification that the simulation \
        has started.

    .. note::
        The machine description database reader can only be used while the
        registered database callbacks are running.
    """

    __slots__ = [
        "__database_callbacks",
        "__pause_and_stop_callback",
        "__running",
        "__start_resume_callback"]

    def __init__(self, start_resume_callback_function=None,
                 stop_pause_callback_function=None, local_host=None,
                 local_port=NOTIFY_PORT):
        """
        :param callable start_resume_callback_function:
            A function to be called when the start message has been received.
            This function should not take any parameters or return anything.
        :param str local_host:
            Optional specification of the local hostname or IP address of the
            interface to listen on
        :param int local_port:
            Optional specification of the local port to listen on. Must match
            the port that the toolchain will send the notification on (19999
            by default)
        """
        super().__init__(
            local_host=local_host, local_port=local_port,
            remote_host=None, remote_port=None)
        thread = Thread(name="SpyNNakerDatabaseConnection:{}:{}".format(
            self.local_ip_address, self.local_port), target=self.__run)
        self.__database_callbacks = list()
        self.__start_resume_callback = start_resume_callback_function
        self.__pause_and_stop_callback = stop_pause_callback_function
        self.__running = False
        thread.daemon = True
        thread.start()

    def add_database_callback(self, database_callback_function):
        """ Add a database callback to be called when the database is ready.

        :param callable(DatabaseReader,None) database_callback_function:
            A function to be called when the database message has been
            received.  This function should take a single parameter, which
            will be a DatabaseReader object. Once the function returns, it
            will be assumed that the database has been read and will not be
            needed further, and the return response will be sent.
        """
        self.__database_callbacks.append(database_callback_function)

    def __run(self):
        # pylint: disable=broad-except
        self.__running = True
        logger.info(
            "{}:{} Waiting for message to indicate that the database is "
            "ready", self.local_ip_address, self.local_port)
        try:
            while self.__running:
                try:
                    data, address = self.receive_with_address(timeout=3)
                except SpinnmanTimeoutException:
                    continue
                self.__read_db(address, data)

                # Wait for the start of the simulation
                if self.__start_resume_callback is not None:
                    self.__start_resume()

                # Wait for the end of the simulation
                if self.__pause_and_stop_callback is not None:
                    self.__pause_stop()
        except Exception as e:
            logger.error("Failure processing database callback",
                         exc_info=True)
            raise SpinnmanIOException(str(e)) from e
        finally:
            self.__running = False

    def __read_db(self, address, data):
        # Read the read packet confirmation
        logger.info("{}:{} Reading database",
                    self.local_ip_address, self.local_port)
        if len(data) > 2:
            database_path = data[2:].decode('utf-8')
            logger.info("database is at {}", database_path)

            # Call the callback
            with DatabaseReader(database_path) as db_reader:
                for db_callback in self.__database_callbacks:
                    db_callback(db_reader)
        else:
            logger.warning("Database path was empty - assuming no database")

        # Send the response
        logger.info("Notifying the toolchain that the database has been read")
        self._send_command(CMDS.DATABASE_CONFIRMATION, address)

    def __start_resume(self):
        logger.info(
            "Waiting for message to indicate that the simulation has "
            "started or resumed")
        command_code = self._receive_command()
        if command_code != CMDS.START_RESUME_NOTIFICATION.value:
            raise SpinnmanInvalidPacketException(
                "command_code",
                "expected a start/resume command code now, and did not "
                "receive it")
        # Call the callback
        self.__start_resume_callback()

    def __pause_stop(self):
        logger.info(
            "Waiting for message to indicate that the simulation has "
            "stopped or paused")
        command_code = self._receive_command()
        if command_code != CMDS.STOP_PAUSE_NOTIFICATION.value:
            raise SpinnmanInvalidPacketException(
                "command_code",
                "expected a pause/stop command code now, and did not "
                "receive it")
        # Call the callback
        self.__pause_and_stop_callback()

    def _send_command(self, command, address):
        self.send_to(EIEIOCommandHeader(command.value).bytestring, address)

    def _receive_command(self):
        return EIEIOCommandHeader.from_bytestring(self.receive(), 0).command

    def close(self):
        self.__running = False
        super().close()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.utilities.database import DatabaseWriter

logger = FormatAdapter(logging.getLogger(__name__))


class DatabaseInterface(object):
    """ Writes a database of the graph(s) and other information.
    """

    __slots__ = [
        # the database writer object
        "_writer",

        # True if the end user has asked for the database to be written
        "_user_create_database",

        # True if the network is computed to need the database to be written
        "_needs_db"
    ]

    def __init__(self):
        self._writer = None
        self._user_create_database = None
        self._needs_db = None

    def __call__(
            self, machine_graph, user_create_database, tags,
            runtime, machine, data_n_timesteps, time_scale_factor,
            machine_time_step, placements, routing_infos, router_tables,
            report_folder, create_atom_to_event_id_mapping=False,
            application_graph=None):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        :param str user_create_database:
        :param ~pacman.model.tags.Tags tags:
        :param int runtime:
        :param ~spinn_machine.Machine machine:
        :param int data_n_timesteps:
        :param int time_scale_factor:
        :param int machine_time_step:
        :param ~pacman.model.placements.Placements placements:
        :param ~pacman.model.routing_info.RoutingInfo routing_infos:
        :param router_tables:
        :type router_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param str report_folder: Where the database will be put.
        :param bool create_atom_to_event_id_mapping:
        :param application_graph:
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph
        :return: Database interface, where the database is located
        :rtype: tuple(DatabaseInterface, str)
        """
        # pylint: disable=too-many-arguments

        self._writer = DatabaseWriter(report_folder)
        self._user_create_database = user_create_database
        # add database generation if requested
        self._needs_db = self._writer.auto_detect_database(machine_graph)

        if self.needs_database:
            logger.info("creating live event connection database in {}",
                        self._writer.database_path)
            self._write_to_db(machine, time_scale_factor, machine_time_step,
                              runtime, application_graph, machine_graph,
                              data_n_timesteps, placements,
                              routing_infos, router_tables, tags,
                              create_atom_to_event_id_mapping)

        return self, self.database_file_path

    @property
    def needs_database(self):
        """
        :rtype: bool
        """
        if self._user_create_database == "None":
            return self._needs_db
        return self._user_create_database == "True"

    @property
    def database_file_path(self):
        """
        :rtype: str or None
        """
        if self.needs_database:
            return self._writer.database_path
        return None

    def _write_to_db(
            self, machine, time_scale_factor, machine_time_step,
            runtime, app_graph, machine_graph, data_n_timesteps,
            placements, routing_infos, router_tables, tags, create_mapping):
        """
        :param ~.Machine machine:
        :param int time_scale_factor:
        :param int machine_time_step:
        :param int runtime:
        :param ~.ApplicationGraph app_graph:
        :param ~.MachineGraph machine_graph:
        :param int data_n_timesteps:
            The number of timesteps for which data space will been reserved
        :param ~.Placements placements:
        :param ~.RoutingInfo routing_infos:
        :param ~.MulticastRoutingTables router_tables:
        :param ~.Tags tags:
        :param bool create_mapping:
        """
        # pylint: disable=too-many-arguments

        with self._writer as w, ProgressBar(
                9, "Creating graph description database") as p:
            w.add_system_params(time_scale_factor, machine_time_step, runtime)
            p.update()
            w.add_machine_objects(machine)
            p.update()
            if app_graph is not None and app_graph.n_vertices:
                w.add_application_vertices(app_graph)
            p.update()
            w.add_vertices(machine_graph, data_n_timesteps, app_graph)
            p.update()
            w.add_placements(placements)
            p.update()
            w.add_routing_infos(routing_infos, machine_graph)
            p.update()
            w.add_routing_tables(router_tables)
            p.update()
            w.add_tags(machine_graph, tags)
            p.update()
            if app_graph is not None and create_mapping:
                w.create_atom_to_event_id_mapping(
                    application_graph=app_graph, machine_graph=machine_graph,
                    routing_infos=routing_infos)
            p.update()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import re
from spinn_utilities.ordered_set import OrderedSet
from spinn_front_end_common.utilities.utility_objs import PowerUsed
from spinn_front_end_common.utility_models import (
    ChipPowerMonitorMachineVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.helpful_functions import (
    convert_time_diff_to_total_milliseconds)
from spinn_front_end_common.utilities.globals_variables import get_simulator

#: milliseconds per second
_MS_PER_SECOND = 1000.0


class ComputeEnergyUsed(object):
    """ This algorithm does the actual work of computing energy used by a\
        simulation (or other application) running on SpiNNaker.
    """

    #: given from Indar's measurements
    MILLIWATTS_PER_FPGA = 0.000584635

    #: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    #: Massively-Parallel Neural Network Simulation)
    JOULES_PER_SPIKE = 0.000000000800

    #: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    #: Massively-Parallel Neural Network Simulation)
    MILLIWATTS_PER_IDLE_CHIP = 0.000360

    #: stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    #: Massively-Parallel Neural Network Simulation)
    MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD = 0.001 - MILLIWATTS_PER_IDLE_CHIP

    #: measured from the real power meter and timing between
    #: the photos for a days powered off
    MILLIWATTS_FOR_FRAME_IDLE_COST = 0.117

    #: measured from the loading of the column and extrapolated
    MILLIWATTS_PER_FRAME_ACTIVE_COST = 0.154163558

    #: measured from the real power meter and timing between the photos
    #: for a day powered off
    MILLIWATTS_FOR_BOXED_48_CHIP_FRAME_IDLE_COST = 0.0045833333

    # TODO needs filling in
    MILLIWATTS_PER_UNBOXED_48_CHIP_FRAME_IDLE_COST = 0.01666667

    # TODO verify this is correct when doing multiboard comms
    N_MONITORS_ACTIVE_DURING_COMMS = 2

    __slots__ = []

    def __call__(
            self, placements, machine, version, time_scale_factor,
            router_provenance, runtime, buffer_manager, mapping_time,
            load_time, execute_time, dsg_time, extraction_time,
            spalloc_server=None, remote_spinnaker_url=None,
            machine_allocation_controller=None):
        """
        :param ~pacman.model.placements.Placements placements:
        :param ~spinn_machine.Machine machine:
        :param int version:
            The version of the SpiNNaker boards in use.
        :param int time_scale_factor:
        :param list(ProvenanceDataItem) router_provenance:
            Provenance information from routers.
        :param float runtime:
        :param BufferManager buffer_manager:
        :param float mapping_time:
            From simulator via :py:class:`~.FinaliseTimingData`.
        :param float load_time:
            From simulator via :py:class:`~.FinaliseTimingData`.
        :param float execute_time:
            From simulator via :py:class:`~.FinaliseTimingData`.
        :param float dsg_time:
            From simulator via :py:class:`~.FinaliseTimingData`.
        :param float extraction_time:
            From simulator via :py:class:`~.FinaliseTimingData`.
        :param spalloc_server: (optional)
        :type spalloc_server: str or None
        :param remote_spinnaker_url: (optional)
        :type remote_spinnaker_url: str or None
        :param MachineAllocationController machine_allocation_controller:
            (optional)
        :rtype: PowerUsed
        """
        # pylint: disable=too-many-arguments

        power_used = PowerUsed()

        # CHEAT! We call into the guts of AbstractSpinnakerBase to get the
        # algorithm timing data; it's still being written to at this point...
        pacman_provenance = get_simulator()._pacman_provenance.data_items
        # Take a copy of the PACMAN algorithm timing provenance data
        power_used._algorithm_timing_provenance = list(pacman_provenance)

        power_used.num_chips = machine.n_chips
        # One extra per chip for SCAMP
        power_used.num_cores = placements.n_placements + machine.n_chips
        power_used.exec_time_secs = execute_time / _MS_PER_SECOND
        power_used.loading_time_secs = load_time / _MS_PER_SECOND
        power_used.saving_time_secs = extraction_time / _MS_PER_SECOND
        power_used.data_gen_time_secs = dsg_time / _MS_PER_SECOND
        power_used.mapping_time_secs = mapping_time / _MS_PER_SECOND

        using_spalloc = bool(spalloc_server or remote_spinnaker_url)
        self._compute_energy_consumption(
             placements, machine, version, using_spalloc,
             router_provenance, dsg_time, buffer_manager, load_time,
             mapping_time, execute_time + load_time + extraction_time,
             machine_allocation_controller,
             runtime * time_scale_factor, power_used)

        return power_used

    def _compute_energy_consumption(
            self, placements, machine, version, using_spalloc,
            router_provenance, dsg_time, buffer_manager, load_time,
            mapping_time, total_booted_time, job, runtime_total_ms,
            power_used):
        """
        :param ~.Placements placements:
        :param ~.Machine machine:
        :param int version:
        :param bool using_spalloc:
        :param list(ProvenanceDataItem) router_provenance:
        :param float dsg_time:
        :param BufferManager buffer_manager:
        :param float load_time:
        :param float mapping_time:
        :param float total_booted_time:
        :param MachineAllocationController job:
        :param float runtime_total_ms:
        :param PowerUsed power_used:
        """
        # figure active chips
        active_chips = self.__active_chips(machine, placements)

        # figure out packet cost
        self._router_packet_energy(router_provenance, power_used)

        # figure FPGA cost over all booted and during runtime cost
        self._calculate_fpga_energy(
            machine, version, using_spalloc, total_booted_time,
            runtime_total_ms, power_used)

        # figure how many frames are using, as this is a constant cost of
        # routers, cooling etc
        power_used.num_frames = self._calculate_n_frames(machine, job)

        # figure load time cost
        power_used.loading_joules = self._calculate_loading_energy(
            power_used._algorithm_timing_provenance, machine, load_time,
            active_chips, power_used.num_frames)

        # figure the down time idle cost for mapping
        power_used.mapping_joules = self._calculate_power_down_energy(
            mapping_time, machine, job, version, power_used.num_frames)

        # figure the down time idle cost for DSG
        power_used.data_gen_joules = self._calculate_power_down_energy(
            dsg_time, machine, job, version, power_used.num_frames)

        # figure extraction time cost
        power_used.saving_joules = self._calculate_data_extraction_energy(
            power_used._algorithm_timing_provenance, machine, active_chips,
            power_used.num_frames)

        # figure out active chips cost
        power_used.chip_energy_joules = sum(
            self._calculate_chips_active_energy(
                chip, placements, buffer_manager, runtime_total_ms, power_used)
            for chip in active_chips)

        # figure out cooling/internet router idle cost during runtime
        power_used.baseline_joules = (
            runtime_total_ms * power_used.num_frames *
            self.MILLIWATTS_FOR_FRAME_IDLE_COST)

    @staticmethod
    def __active_chips(machine, placements):
        """
        :param ~.Machine machine:
        :param ~.Placements placements
        :rtype: set(~.Chip)
        """
        return OrderedSet(
            machine.get_chip_at(placement.x, placement.y)
            for placement in placements
            if isinstance(placement.vertex, ChipPowerMonitorMachineVertex))

    _PER_CHIP_NAMES = frozenset((
        "expected_routers", "unexpected_routers"))
    _MULTICAST_COUNTER_NAMES = frozenset((
        "Local_Multicast_Packets", "External_Multicast_Packets", "Reinjected"))
    _PEER_TO_PEER_COUNTER_NAMES = frozenset((
        "Local_P2P_Packets", "External_P2P_Packets"))
    _NEAREST_NEIGHBOUR_COUNTER_NAMES = frozenset((
        "Local_NN_Packets", "External_NN_Packets"))
    _FIXED_ROUTE_COUNTER_NAMES = frozenset((
        "Local_FR_Packets", "External_FR_Packets"))

    def _router_packet_energy(self, router_provenance, power_used):
        """
        :param list(ProvenanceDataItem) router_provenance:
        :param PowerUsed power_used:
        """
        energy_cost = 0.0
        for element in router_provenance:
            # only process per chip counters, not summary counters.
            if element.names[1] not in self._PER_CHIP_NAMES:
                continue

            # process MC packets
            if element.names[3] in self._MULTICAST_COUNTER_NAMES:
                this_cost = float(element.value) * self.JOULES_PER_SPIKE

            # process p2p packets
            elif element.names[3] in self._PEER_TO_PEER_COUNTER_NAMES:
                this_cost = float(element.value) * self.JOULES_PER_SPIKE * 2

            # process NN packets
            elif element.names[3] in self._NEAREST_NEIGHBOUR_COUNTER_NAMES:
                this_cost = float(element.value) * self.JOULES_PER_SPIKE

            # process FR packets
            elif element.names[3] in self._FIXED_ROUTE_COUNTER_NAMES:
                this_cost = float(element.value) * self.JOULES_PER_SPIKE * 2

            else:
                # ???
                this_cost = 0.0

            energy_cost += this_cost

            # If we can record this against a particular chip, also do so
            m = re.match(r"router_at_chip_(\d+)_(\d+)", element.names[2])
            if m and this_cost:
                x = int(m.group(1))
                y = int(m.group(2))
                power_used.add_router_active_energy(x, y, this_cost)

        power_used.packet_joules = energy_cost

    def _calculate_chips_active_energy(
            self, chip, placements, buffer_manager, runtime_total_ms,
            power_used):
        """ Figure out the chip active cost during simulation

        :param ~.Chip chip: the chip to consider
        :param ~.Placements placements: placements
        :param BufferManager buffer_manager: buffer manager
        :param float runtime_total_ms:
        :param PowerUsed power_used:
        :return: energy cost
        """
        # pylint: disable=too-many-arguments

        # locate chip power monitor
        chip_power_monitor = self.__get_chip_power_monitor(chip, placements)

        # get recordings from the chip power monitor
        recorded_measurements = chip_power_monitor.get_recorded_data(
            placement=placements.get_placement_of_vertex(chip_power_monitor),
            buffer_manager=buffer_manager)

        # deduce time in milliseconds per recording element
        time_for_recorded_sample = (
            chip_power_monitor.sampling_frequency *
            chip_power_monitor.n_samples_per_recording) / 1000
        cores_power_cost = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        # accumulate costs
        for recorded_measurement in recorded_measurements:
            for core in range(0, 18):
                cores_power_cost[core] += (
                    recorded_measurement[core] * time_for_recorded_sample *
                    self.MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD / 18)

        # detailed report print out
        for core in range(0, 18):
            power_used.add_core_active_energy(
                chip.x, chip.y, core, cores_power_cost[core])

        # TAKE INTO ACCOUNT IDLE COST
        idle_cost = runtime_total_ms * self.MILLIWATTS_PER_IDLE_CHIP
        return sum(cores_power_cost) + idle_cost

    def __get_chip_power_monitor(self, chip, placements):
        """ Locate chip power monitor

        :param ~.Chip chip: the chip to consider
        :param ~.Placements placements: placements
        :return: the machine vertex coupled to the monitor
        :rtype: ChipPowerMonitorMachineVertex
        :raises Exception: if it can't find the monitor
        """
        # TODO this should be in the ChipPowerMonitor class
        # it is its responsibility, but needs the self-partitioning

        # start at top, as more likely it was placed on the top
        for processor_id in range(chip.n_processors):
            if placements.is_processor_occupied(chip.x, chip.y, processor_id):
                # check if vertex is a chip power monitor
                vertex = placements.get_vertex_on_processor(
                    chip.x, chip.y, processor_id)
                if isinstance(vertex, ChipPowerMonitorMachineVertex):
                    return vertex

        raise Exception("expected to find a chip power monitor!")

    def _calculate_fpga_energy(
            self, machine, version, using_spalloc, total_runtime,
            runtime_total_ms, power_used):
        """
        :param ~.Machine machine:
        :param int version:
        :param bool using_spalloc:
        :param float total_runtime:
        :param float runtime_total_ms:
        :param PowerUsed power_used:
        """
        # pylint: disable=too-many-arguments

        total_fpgas = 0
        # if not spalloc, then could be any type of board
        if not using_spalloc:
            # if a spinn2 or spinn3 (4 chip boards) then they have no fpgas
            if int(version) in (2, 3):
                return 0, 0
            elif int(version) not in (4, 5):
                # No idea what we've got here!
                raise ConfigurationException(
                    "Do not know what the FPGA setup is for this version of "
                    "SpiNNaker machine.")

            # if the spinn4 or spinn5 board, need to verify if wrap-arounds
            # are there, if not then assume fpgas are turned off.

            # how many fpgas are active
            total_fpgas = self.__board_n_operational_fpgas(
                machine, machine.ethernet_connected_chips[0])
            # active fpgas
            if total_fpgas == 0:
                return 0, 0
        else:  # spalloc machine, need to check each board
            for ethernet_connected_chip in machine.ethernet_connected_chips:
                total_fpgas += self.__board_n_operational_fpgas(
                    machine, ethernet_connected_chip)

        # Only need to update this here now that we've learned there are FPGAs
        # in use
        power_used.num_fpgas = total_fpgas
        power_usage_total = (
            total_runtime * self.MILLIWATTS_PER_FPGA * total_fpgas)
        power_usage_runtime = (
            runtime_total_ms * self.MILLIWATTS_PER_FPGA * total_fpgas)
        power_used.fpga_total_energy_joules = power_usage_total
        power_used.fpga_exec_energy_joules = power_usage_runtime

    def __board_n_operational_fpgas(self, machine, ethernet_chip):
        """ Figures out how many FPGAs were switched on for a particular \
            SpiNN-5 board.

        :param ~.Machine machine: SpiNNaker machine
        :param ~.Chip ethernet_chip: the ethernet chip to look from
        :return: number of FPGAs on, on this board
        """
        # pylint: disable=too-many-locals

        # TODO: should be possible to get this info from Machine

        # positions to check for active links
        left_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((0, 0), (0, 1), (0, 2), (0, 3), (0, 4)))
        right_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((7, 3), (7, 4), (7, 5), (7, 6), (7, 7)))
        top_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((4, 7), (5, 7), (6, 7), (7, 7)))
        bottom_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((0, 0), (1, 0), (2, 0), (3, 0), (4, 0)))
        top_left_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((0, 3), (1, 4), (2, 5), (3, 6), (4, 7)))
        bottom_right_chips = (
            machine.get_chip_at(ethernet_chip.x + dx, ethernet_chip.y + dy)
            for dx, dy in ((0, 4), (1, 5), (2, 6), (3, 7)))

        # bottom left, bottom
        fpga_0 = self.__deduce_fpga(
            bottom_chips, bottom_right_chips, (5, 4), (0, 5))
        # left, and top right
        fpga_1 = self.__deduce_fpga(
            left_chips, top_left_chips, (3, 4), (3, 2))
        # top and right
        fpga_2 = self.__deduce_fpga(
            top_chips, right_chips, (2, 1), (0, 1))
        return fpga_0 + fpga_1 + fpga_2

    @staticmethod
    def __deduce_fpga(chips_1, chips_2, links_1, links_2):
        """ Figure out if each FPGA was on or not

        :param iterable(~.Chip) chips_1: chips on an edge of the board
        :param iterable(~.Chip) chips_2: chips on an edge of the board
        :param iterable(int) links_1: which link IDs to check from chips_1
        :param iterable(int) links_2: which link IDs to check from chips_2
        :return: 0 if not on, 1 if on
        :rtype: int
        """
        for chip, link_id in itertools.product(chips_1, links_1):
            if chip and chip.router.get_link(link_id) is not None:
                return 1
        for chip, link_id in itertools.product(chips_2, links_2):
            if chip and chip.router.get_link(link_id) is not None:
                return 1
        return 0

    @staticmethod
    def __get_running_time_ms(pacman_provenance, filter_rule):
        return float(sum(
            convert_time_diff_to_total_milliseconds(element.value)
            for element in pacman_provenance
            if filter_rule(element.names)))

    def _calculate_loading_energy(
            self, pacman_provenance, machine, load_time_ms, active_chips,
            n_frames):
        """
        :param list(ProvenanceDataItem) pacman_provenance:
        :param ~.Machine machine:
        :param float load_time_ms: milliseconds
        :param list active_chips:
        :param int n_frames:
        :rtype: float
        """
        # pylint: disable=too-many-arguments

        # find time in milliseconds
        total_time_ms = self.__get_running_time_ms(
            pacman_provenance, lambda names: names[1] == "loading")

        # handle monitor core active cost

        # min between chips that are active and fixed monitor, as when 1
        # chip is used its one monitor, if more than 1 chip,
        # the ethernet connected chip and the monitor handling the read/write
        # this is checked by min
        n_monitors_active = min(
            self.N_MONITORS_ACTIVE_DURING_COMMS, len(active_chips))
        energy_cost = (
            total_time_ms * n_monitors_active *
            self.MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD /
            machine.DEFAULT_MAX_CORES_PER_CHIP)

        # handle all idle cores
        energy_cost += self._calculate_idle_cost(total_time_ms, machine)

        # handle time diff between load time and total load phase of ASB
        energy_cost += (
            (load_time_ms - total_time_ms) *
            machine.n_chips * self.MILLIWATTS_PER_IDLE_CHIP)

        # handle active routers etc
        active_router_cost = (
            load_time_ms * n_frames * self.MILLIWATTS_PER_FRAME_ACTIVE_COST)

        # accumulate
        energy_cost += active_router_cost
        return energy_cost

    def _calculate_data_extraction_energy(
            self, pacman_provenance, machine, active_chips, n_frames):
        """ Data extraction cost

        :param list(ProvenanceDataItem) pacman_provenance:
            provenance items from the PACMAN set
        :param ~.Machine machine: machine description
        :param list active_chips:
        :param int n_frames:
        :return: cost of data extraction in Joules
        :rtype: float
        """
        # pylint: disable=too-many-arguments

        # find time
        total_time_ms = self.__get_running_time_ms(
            pacman_provenance, lambda names: (
                names[1] == "Execution" and names[2] !=
                "run_time_of_FrontEndCommonApplicationRunner"))

        # min between chips that are active and fixed monitor, as when 1
        # chip is used its one monitor, if more than 1 chip,
        # the ethernet connected chip and the monitor handling the read/write
        # this is checked by min
        energy_cost = (
            total_time_ms *
            min(self.N_MONITORS_ACTIVE_DURING_COMMS, len(active_chips)) *
            self.MILLIWATTS_PER_CHIP_ACTIVE_OVERHEAD /
            machine.DEFAULT_MAX_CORES_PER_CHIP)

        # add idle chip cost
        energy_cost += self._calculate_idle_cost(total_time_ms, machine)

        # handle active routers etc
        energy_cost_of_active_router = (
            total_time_ms * n_frames * self.MILLIWATTS_PER_FRAME_ACTIVE_COST)
        energy_cost += energy_cost_of_active_router
        return energy_cost

    @classmethod
    def _calculate_idle_cost(cls, time, machine):
        """ Calculate energy used by being idle.

        :param float time: time machine was idle, in milliseconds
        :param ~.Machine machine: machine description
        :return: cost in joules
        :rtype: float
        """
        return (time * machine.total_available_user_cores *
                cls.MILLIWATTS_PER_IDLE_CHIP /
                machine.DEFAULT_MAX_CORES_PER_CHIP)

    @classmethod
    def _calculate_power_down_energy(
            cls, time, machine, job, version, n_frames):
        """ Calculate power down costs

        :param float time: time powered down, in milliseconds
        :param ~.Machine machine:
        :param AbstractMachineAllocationController job:
            the spalloc job object
        :param int version:
        :param int n_frames: number of frames used by this machine
        :return: energy in joules
        :rtype: float
        """
        # pylint: disable=too-many-arguments

        # if spalloc or hbp
        if job is not None:
            return time * n_frames * cls.MILLIWATTS_FOR_FRAME_IDLE_COST
        # if 48 chip
        elif version == 5 or version == 4:
            return time * cls.MILLIWATTS_FOR_BOXED_48_CHIP_FRAME_IDLE_COST
        # if 4 chip
        elif version == 3 or version == 2:
            return machine.n_chips * time * cls.MILLIWATTS_PER_IDLE_CHIP
        # boom
        else:
            raise ConfigurationException("don't know what to do here")

    @staticmethod
    def _calculate_n_frames(machine, job):
        """ Figures out how many frames are being used in this setup.\
            A key of cabinet,frame will be used to identify unique frame.

        :param ~.Machine machine: the machine object
        :param AbstractMachineAllocationController job:
            the spalloc job object
        :return: number of frames
        :rtype: int
        """

        # if not spalloc, then could be any type of board, but unknown cooling
        if job is None:
            return 0

        # if using spalloc in some form
        cabinet_frame = set()
        for ethernet_connected_chip in machine.ethernet_connected_chips:
            cabinet, frame, _ = job.where_is_machine(
                ethernet_connected_chip.x, ethernet_connected_chip.y)
            cabinet_frame.add((cabinet, frame))
        return len(cabinet_frame)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
from collections.abc import MutableMapping
from .data_row_writer import DataRowWriter
from .ds_sqllite_database import DsSqlliteDatabase


class DataSpecificationTargets(MutableMapping):

    __slots__ = ["_db"]

    def __init__(self, machine, report_folder, init=None, clear=True):
        """
        :param ~spinn_machine.Machine machine:
        :param str report_folder:
        :param init:
        :type init: bool or None
        :param bool clear:
        """
        # pylint: disable=super-init-not-called
        # real DB would write to report_folder
        self._db = DsSqlliteDatabase(machine, report_folder, init)
        # Clear any current data specs
        if clear:
            self._db.clear_ds()

    def __getitem__(self, core):
        """ Implements the mapping `__getitem__` as long as core is the right
            type.

        :param tuple(int,int,int) core: triple of (x, y, p)
        :return: dictionary with the keys ``start_address``, ``memory_used``
            and ``memory_written``
        :rtype: ~io.RawIOBase
        """
        (x, y, p) = core
        return io.BytesIO(self._db.get_ds(x, y, p))

    def __setitem__(self, core, info):
        raise NotImplementedError(
            "Direct set not supported. See create_data_spec")

    def __delitem__(self, core):
        raise NotImplementedError("Delete not supported")

    def keys(self):
        """ Yields the keys.

        As the more typical call is iteritems this makes use of that

        :rtype: iterable(tuple(int,int,int))
        """
        for key, _value in self._db.ds_iteritems():
            yield key

    __iter__ = keys

    def __len__(self):
        """
        TEMP implementation

        :return:
        """
        return self._db.ds_n_cores()

    n_targets = __len__

    def create_data_spec(self, x, y, p):
        """
        :param int x:
        :param int y:
        :param int p:
        :rtype: DataRowWriter
        """
        return DataRowWriter(x, y, p, self)

    def write_data_spec(self, x, y, p, ds):
        """
        :param int x:
        :param int y:
        :param int p:
        :param bytearray ds:
        """
        self._db.save_ds(x, y, p, ds)

    def items(self):
        """
        :return: iterator over the core locations and how to read the data
            spec for them
        :rtype: iterable(tuple(tuple(int,int,int),~io.RawIOBase))
        """
        for key, value in self._db.ds_iteritems():
            yield key, io.BytesIO(value)

    def get_database(self):
        """ Expose the database so it can be shared

        :rtype: DsAbstractDatabase
        """
        return self._db

    def set_app_id(self, app_id):
        """ Sets the same app_id for all rows that have DS content

        :param int app_id: value to set
        """
        self._db.ds_set_app_id(app_id)

    def get_app_id(self, x, y, p):
        """ Gets the app_id set for this core

        :param int x: core x
        :param int y: core y
        :param int p: core p
        :rtype: int
        """
        return self._db.ds_get_app_id(x, y, p)

    def mark_system_cores(self, core_subsets):
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
        """
        cores_to_mark = []
        for subset in core_subsets:
            x = subset.x
            y = subset.y
            for p in subset.processor_ids:
                cores_to_mark.append((x, y, p))
        self._db.ds_mark_as_system(cores_to_mark)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractMachineAllocationController(object, metaclass=AbstractBase):
    """ An object that controls the allocation of a machine
    """

    __slots__ = ()

    @abstractmethod
    def extend_allocation(self, new_total_run_time):
        """ Extend the allocation of the machine from the original run time.

        :param float new_total_run_time:
            The total run time that is now required starting from when the
            machine was first allocated
        """

    @abstractmethod
    def close(self):
        """ Indicate that the use of the machine is complete.
        """

    @abstractmethod
    def where_is_machine(self, chip_x, chip_y):
        """ Locates and returns cabinet, frame, board for a given chip in a\
            machine allocated to this job.

        :param int chip_x: chip x location
        :param int chip_y: chip y location
        :return: (cabinet, frame, board)
        :rtype: tuple(int,int,int)
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.log import FormatAdapter
from spinn_machine import json_machine, virtual_machine, Machine, Router

logger = FormatAdapter(logging.getLogger(__name__))


class VirtualMachineGenerator(object):
    """ Generates a virtual machine with given dimensions and configuration.
    """

    __slots__ = []

    def __call__(
            self, width=None, height=None, version=None,
            down_chips=None, down_cores=None, down_links=None,
            max_sdram_size=None,
            router_entries_per_chip=Router.ROUTER_DEFAULT_AVAILABLE_ENTRIES,
            json_path=None):
        """
        :param int width: The width of the machine in chips
        :param int height: The height of the machine in chips
        :param int version: The version of board to create
        :param list(tuple(int,int)) down_chips:
            The set of chips that should be considered broken
        :param list(tuple(int,int,int)) down_cores:
            The set of cores that should be considered broken
        :param list(tuple(int,int,int)) down_links:
            The set of links that should be considered broken
        :param int max_sdram_size: The SDRAM that should be given to each chip
        :param int router_entries_per_chip:
            The number of router entries to allocate.
        :param str json_path:
            Where to load a JSON description of the machine from, if anywhere.
        :return: The virtual machine.
        :rtype: ~spinn_machine.Machine
        :raises Exception: If given bad arguments
        """
        # For backward compatibility support version in csf files for now
        if version is not None:
            if version in [2, 3]:
                if height is None:
                    height = 2
                else:
                    assert height == 2
                if width is None:
                    width = 2
                else:
                    assert width == 2
                logger.warning("For virtual Machines version is deprecated."
                               "use width=2, height=2 instead")
            elif version in [4, 5]:
                if height is None:
                    height = 8
                else:
                    assert height == 8
                if width is None:
                    width = 8
                else:
                    assert width == 8
                logger.warning("For virtual Machines version is deprecated."
                               "use width=8, height=8 instead")
            else:
                raise Exception("Unknown version {}".format(version))

        if json_path is None:
            # pylint: disable=too-many-arguments
            machine = virtual_machine(
                width=width, height=height,
                n_cpus_per_chip=Machine.max_cores_per_chip(),
                down_chips=down_chips, down_cores=down_cores,
                down_links=down_links, sdram_per_chip=max_sdram_size,
                router_entries_per_chip=router_entries_per_chip, validate=True)
        else:
            if (height is not None or width is not None or
                    version is not None or down_chips is not None or
                    down_cores is not None or down_links is not None):
                logger.warning("As json_path specified all other virtual "
                               "machine settings ignored.")
            machine = json_machine.machine_from_json(json_path)

        # Work out and add the SpiNNaker links and FPGA links
        machine.add_spinnaker_links()
        machine.add_fpga_links()

        logger.info(
            "Created a virtual machine which has {}".format(
                machine.cores_and_link_output_string()))

        return machine



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB


class DatabaseReader(SQLiteDB):
    """ A reader for the database.
    """
    __slots__ = []

    def __init__(self, database_path):
        """
        :param str database_path: The path to the database
        """
        super().__init__(database_path, read_only=True, text_factory=str)

    def __exec_one(self, query, *args):
        with self.transaction() as cur:
            cur.execute(query + " LIMIT 1", args)
            return cur.fetchone()

    @staticmethod
    def __r2t(row, *args):
        return tuple(None if row is None else row[key] for key in args)

    def get_key_to_atom_id_mapping(self, label):
        """ Get a mapping of event key to atom ID for a given vertex

        :param str label: The label of the vertex
        :return: dictionary of atom IDs indexed by event key
        :rtype: dict(int, int)
        """
        with self.transaction() as cur:
            return {
                row["event"]: row["atom"]
                for row in cur.execute(
                    """
                    SELECT * FROM label_event_atom_view
                    WHERE label = ?
                    """, (label, ))}

    def get_atom_id_to_key_mapping(self, label):
        """ Get a mapping of atom ID to event key for a given vertex

        :param str label: The label of the vertex
        :return: dictionary of event keys indexed by atom ID
        :rtype: dict(int, int)
        """
        with self.transaction() as cur:
            return {
                row["atom"]: row["event"]
                for row in cur.execute(
                    """
                    SELECT * FROM label_event_atom_view
                    WHERE label = ?
                    """, (label, ))}

    def get_live_output_details(self, label, receiver_label):
        """ Get the IP address, port and whether the SDP headers are to be\
            stripped from the output from a vertex

        :param str label: The label of the vertex
        :return: tuple of (IP address, port, strip SDP, board address, tag)
        :rtype: tuple(str, int, bool, str, int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM app_output_tag_view
                WHERE pre_vertex_label = ?
                    AND post_vertex_label LIKE ?
                """, label, str(receiver_label) + "%"),
            "ip_address", "port", "strip_sdp", "board_address", "tag")

    def get_live_input_details(self, label):
        """ Get the IP address and port where live input should be sent\
            for a given vertex

        :param str label: The label of the vertex
        :return: tuple of (IP address, port)
        :rtype: tuple(str, int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM app_input_tag_view
                WHERE application_label = ?
                """, label),
            "board_address", "port")

    def get_machine_live_output_details(self, label, receiver_label):
        """ Get the IP address, port and whether the SDP headers are to be\
            stripped from the output from a machine vertex

        :param str label: The label of the vertex
        :param str receiver_label:
        :return: tuple of (IP address, port, strip SDP, board address, tag)
        :rtype: tuple(str, int, bool, str, int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM machine_output_tag_view
                WHERE pre_vertex_label = ?
                    AND post_vertex_label LIKE ?
                """, label, str(receiver_label) + "%"),
            "ip_address", "port", "strip_sdp", "board_address", "tag")

    def get_machine_live_input_details(self, label):
        """ Get the IP address and port where live input should be sent\
            for a given machine vertex

        :param str label: The label of the vertex
        :return: tuple of (IP address, port)
        :rtype: tuple(str, int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM machine_input_tag_view
                WHERE machine_label = ?
                """, label),
            "board_address", "port")

    def get_machine_live_output_key(self, label, receiver_label):
        """
        :param str label: The label of the vertex
        :param str receiver_label:
        :rtype: tuple(int,int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM machine_edge_key_view
                WHERE pre_vertex_label = ?
                    AND post_vertex_label LIKE ?
                """, label, str(receiver_label) + "%"),
            "key", "mask")

    def get_machine_live_input_key(self, label):
        """
        :param str label: The label of the vertex
        :rtype: tuple(int,int)
        """
        return self.__r2t(
            self.__exec_one(
                """
                SELECT * FROM machine_edge_key_view
                WHERE pre_vertex_label = ?
                """, label),
            "key", "mask")

    def get_n_atoms(self, label):
        """ Get the number of atoms in a given vertex

        :param str label: The label of the vertex
        :return: The number of atoms
        :rtype: int
        """
        row = self.__exec_one(
            """
            SELECT no_atoms FROM Application_vertices
            WHERE vertex_label = ?
            """, label)
        return 0 if row is None else row["no_atoms"]

    def get_configuration_parameter_value(self, parameter_name):
        """ Get the value of a configuration parameter

        :param str parameter_name: The name of the parameter
        :return: The value of the parameter
        :rtype: float or None
        """
        row = self.__exec_one(
            """
            SELECT value FROM configuration_parameters
            WHERE parameter_id = ?
            """, parameter_name)
        return None if row is None else float(row["value"])

    @staticmethod
    def __xyp(row):
        return int(row["x"]), int(row["y"]), int(row["p"])

    def get_placement(self, label):
        """ Get the placement of a machine vertex with a given label

        :param str label: The label of the vertex
        :return: The x, y, p coordinates of the vertex
        :rtype: tuple(int, int, int)
        """
        row = self.__exec_one(
            """
            SELECT x, y, p FROM machine_vertex_placement
            WHERE vertex_label = ?
            """, label)
        return (None, None, None) if row is None else self.__xyp(row)

    def get_placements(self, label):
        """ Get the placements of an application vertex with a given label

        :param str label: The label of the vertex
        :return: A list of x, y, p coordinates of the vertices
        :rtype: list(tuple(int, int, int))
        """
        with self.transaction() as cur:
            return [
                self.__xyp(row) for row in cur.execute(
                    """
                    SELECT x, y, p FROM application_vertex_placements
                    WHERE vertex_label = ?
                    """, (label, ))]

    def get_ip_address(self, x, y):
        """ Get an IP address to contact a chip

        :param int x: The x-coordinate of the chip
        :param int y: The y-coordinate of the chip
        :return: The IP address of the Ethernet to use to contact the chip
        :rtype: str or None
        """
        row = self.__exec_one(
            """
            SELECT eth_ip_address FROM chip_eth_info
            WHERE x = ? AND y = ? OR x = 0 AND y = 0
            ORDER BY x DESC
            """, x, y)
        # Should only fail if no machine is present!
        return None if row is None else row["eth_ip_address"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
from spinn_machine import Machine
from spinn_machine.virtual_machine import virtual_machine


class HBPMaxMachineGenerator(object):
    """ Generates a virtual machine of the width and height of the maximum\
        machine a given HBP server can generate.
    """

    __slots__ = []

    def __call__(self, hbp_server_url, total_run_time,
                 max_machine_core_reduction=0):
        """
        :param str hbp_server_url:
            The URL of the HBP server from which to get the machine
        :param int total_run_time: The total run time to request
        :param int max_machine_core_reduction:
            the number of cores less than
            :py:const:`~spinn_machine.Machine.DEFAULT_MAX_CORES_PER_CHIP`
            that each chip should have
        :rtype: ~spinn_machine.Machine
        """

        max_machine = self._max_machine_request(hbp_server_url, total_run_time)

        n_cpus_per_chip = (Machine.max_cores_per_chip() -
                           max_machine_core_reduction)

        # Return the width and height and assume that it has wrap arounds
        return virtual_machine(
            width=max_machine["width"], height=max_machine["height"],
            n_cpus_per_chip=n_cpus_per_chip, validate=False)

    def _max_machine_request(self, url, total_run_time):
        """
        :param str url:
        :param int total_run_time:
        :rtype: dict
        """
        if url.endswith("/"):
            url = url[:-1]
        r = requests.get("{}/max".format(url), params={
            'runTime': total_run_time})
        r.raise_for_status()
        return r.json()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractproperty


class AbstractSendMeMulticastCommandsVertex(object, metaclass=AbstractBase):
    """ A device that may be a virtual vertex which wants to commands to be\
        sent to it as multicast packets at fixed points in the simulation.

    .. note::
        The device might not be a vertex at all. It could instead be
        instantiated entirely host side, in which case these methods will
        never be called.
    """

    __slots__ = ()

    @abstractproperty
    def start_resume_commands(self):
        """ The commands needed when starting or resuming simulation

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """

    @abstractproperty
    def pause_stop_commands(self):
        """ The commands needed when pausing or stopping simulation

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """

    @abstractproperty
    def timed_commands(self):
        """ The commands to be sent at given times in the simulation

        :rtype:
            iterable(~spinn_front_end_common.utility_models.MultiCastCommand)
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from spinn_front_end_common.utilities.notification_protocol import (
    NotificationProtocol)


class CreateNotificationProtocol(object):
    """ Builds the notification protocol for GUI and external device \
        interaction.
    """

    __slots__ = []

    def __call__(
            self, wait_for_read_confirmation,
            socket_addresses, database_file_path):
        """
        :param bool wait_for_read_confirmation:
        :param socket_addresses: Where to notify.
        :type socket_addresses:
            list(~spinn_utilities.socket_address.SocketAddress)
        :param str database_file_path:
        """
        notification_protocol = NotificationProtocol(
            socket_addresses, wait_for_read_confirmation)
        notification_protocol.send_read_notification(database_file_path)
        return notification_protocol



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import traceback
from spinn_utilities.log import FormatAdapter
from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.utility_objs.extra_monitor_scp_messages\
    import (
        GetReinjectionStatusMessage)

logger = FormatAdapter(logging.getLogger(__name__))


class ReadStatusProcess(AbstractMultiConnectionProcess):
    """ How to send messages to read the status of extra monitors.
    """
    __slots__ = ()

    @staticmethod
    def __handle_response(result, response):
        """
        :param dict result:
        :param GetReinjectionStatusMessageResponse response:
        """
        status = response.reinjection_functionality_status
        header = response.sdp_header
        result[header.source_chip_x, header.source_chip_y] = status

    def get_reinjection_status(self, x, y, p):
        """
        :param int x:
        :param int y:
        :param int p:
        :rtype: ReInjectionStatus
        """
        status = dict()
        self._send_request(GetReinjectionStatusMessage(x, y, p),
                           functools.partial(self.__handle_response, status))
        self._finish()
        self.check_for_error()
        return status[x, y]

    def get_reinjection_status_for_core_subsets(self, core_subsets):
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
        :rtype: dict(tuple(int,int), ReInjectionStatus)
        """
        status = dict()
        for core_subset in core_subsets.core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(GetReinjectionStatusMessage(
                    core_subset.x, core_subset.y, processor_id),
                    functools.partial(self.__handle_response, status))
        self._finish()
        if self.is_error():
            logger.warning("Error(s) reading reinjection status:")
            for (e, tb) in zip(self._exceptions, self._tracebacks):
                traceback.print_exception(type(e), e, tb)
        return status



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs import AbstractVertex


@require_subclass(AbstractVertex)
class AbstractProvidesOutgoingPartitionConstraints(
        object, metaclass=AbstractBase):
    """ A vertex that can provide constraints for its outgoing edge partitions.

    If a Machine_vertex is an instance the Application vertex will not be
    checked. However if the MachineVertex does not implement this API
    ProcessPartitionConstraint will then check the ApplicationVertex
    """

    __slots__ = ()

    @abstractmethod
    def get_outgoing_partition_constraints(self, partition):
        """ Get constraints to be added to the given edge partition that comes\
            out of this vertex.

        :param ~pacman.model.graphs.AbstractOutgoingEdgePartition partition:
            An edge that comes out of this vertex
        :return: A list of constraints
        :rtype: list(~pacman.model.constraints.AbstractConstraint)
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
random math constants
"""

# Scaling factors for FP arithmetic, see Xin Jin et al. (2008)
LOG_P1 = 8.0
LOG_P2 = 16.0
P1 = 2.0 ** LOG_P1
P2 = 2.0 ** LOG_P2



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.progress_bar import ProgressBar
from pacman.model.routing_info import DictBasedMachinePartitionNKeysMap
from spinn_front_end_common.utilities.exceptions import ConfigurationException


class EdgeToNKeysMapper(object):
    """ Works out the number of keys needed for each edge.
    """

    __slots__ = []

    PROG_BAR_NAME = (
        "Getting number of keys required by each edge using application graph")
    ERROR_MSG = (
        "A machine graph is required for this mapper. Please choose and try "
        "again")

    def __call__(self, machine_graph):
        """
        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
        :rtype: ~pacman.model.routing_info.DictBasedMachinePartitionNKeysMap
        :raises ConfigurationException: If no graph is available
        """
        if machine_graph is None:
            raise ConfigurationException(self.ERROR_MSG)

        # Generate an n_keys map for the graph and add constraints
        n_keys_map = DictBasedMachinePartitionNKeysMap()

        # generate progress bar
        progress = ProgressBar(machine_graph.n_vertices, self.PROG_BAR_NAME)

        # iterate over each partition in the graph
        for vertex in progress.over(machine_graph.vertices):
            for partition in machine_graph.\
                    get_multicast_edge_partitions_starting_at_vertex(vertex):
                n_keys = partition.pre_vertex.get_n_keys_for_partition(
                    partition)
                n_keys_map.set_n_keys_for_partition(partition, n_keys)

        return n_keys_map



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class SpeedupInSCPCommands(Enum):
    """ SCP Command codes for data speed up in
    """
    SAVE_APPLICATION_MC_ROUTES = 6
    LOAD_APPLICATION_MC_ROUTES = 7
    LOAD_SYSTEM_MC_ROUTES = 8

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access, unused-argument
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2020-2021 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math
from spinn_utilities.log import FormatAdapter
from spinn_front_end_common.abstract_models.impl.\
    tdma_aware_application_vertex import (
        TDMAAwareApplicationVertex)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.helpful_functions import (
    read_config_int, read_config_float)
from spinn_front_end_common.utilities.globals_variables import get_simulator
from spinn_front_end_common.utilities.constants import CLOCKS_PER_US

logger = FormatAdapter(logging.getLogger(__name__))


class LocalTDMABuilder(object):
    """ Builds a localised TDMA which allows a number of machine vertices
    of the same application vertex to fire at the same time. Ensures that
    other application vertices are not firing at the same time. Verifies if
    the total time required fits into the time scale factor and machine time
    step. Below are text diagrams to show how this works in principle.

    *Figure 1:* bits needed to figure out time between spikes.
    Cores 0-4 have 2 atoms, core 5 has 1 atom::

        #        0     1       2      3       4      5
        # T2-[   X                    X
        #    |         X                      X
        #    |                 X                     X
        #    [  X                     X
        #       |------| T
        #              X                      X
        #                      X <- T3

        T = time_between_cores
        T2 = time_between_phases
        T3 = end of TDMA (equiv of ((n_phases + 1) * T2))
        cutoff = 2. n_phases = 3 max_atoms = 2

    Constants etc just to get into head:

    * clock cycles = 200 Mhz = 200 = sv->cpu_clk
    * 1ms = 200000 for timer 1. = clock cycles
    * 200 per microsecond
    * machine time step = microseconds already.
    * `__time_between_cores` = microseconds.

    *Figure 2:* initial offset (used to try to interleave packets from other
    app verts into the TDMA without extending the overall time, and
    trying to stop multiple packets in flight at same time).

    *Figure 3:* bits needed to figure out time between spikes.
    Cores 0-4 have 2 atoms, core 5 has 1 atom::

        #        0  .5   1   .5    2   .5   3    .5   4   .5   5  .5
        # T2-[   X   Y                      X     Y
        #    |           X   Y                        X    Y
        #    |                     X    Y                      X   Y
        #    [  X    Y                      X     Y
        #       |-------| T
        #               X    Y                        X    Y
        #               |----| T4
        #                   T3 ->  X    Y

        T4 is the spreader between populations.
        X is pop0 firing,
        Y is pop1 firing
    """

    # default fraction of the real time we will use for spike transmissions
    FRACTION_OF_TIME_FOR_SPIKE_SENDING = 0.8
    FRACTION_OF_TIME_STEP_BEFORE_SPIKE_SENDING = 0.1

    def __call__(
            self, machine_graph, machine_time_step, time_scale_factor,
            n_keys_map, application_graph=None):
        """ main entrance

        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            machine graph.
        :param int machine_time_step: the machine time step.
        :param int time_scale_factor: the time scale factor.
        :param n_keys_map: the map of partitions to n keys.
        :type n_keys_map:
            ~pacman.model.routing_info.AbstractMachinePartitionNKeysMap
        :param application_graph: app graph.
        :type application_graph:
            ~pacman.model.graphs.application.ApplicationGraph or None
        """
        if application_graph is None:
            return

        # get config params
        us_per_cycle = int(math.ceil(machine_time_step * time_scale_factor))
        clocks_per_cycle = us_per_cycle * CLOCKS_PER_US
        (app_machine_quantity, clocks_between_cores, clocks_for_sending,
         clocks_waiting, clocks_initial) = self.config_values(clocks_per_cycle)

        # calculate for each app vertex if the time needed fits
        app_verts = list()
        max_fraction_of_sending = 0
        for app_vertex in application_graph.vertices:
            if isinstance(app_vertex, TDMAAwareApplicationVertex):
                app_verts.append(app_vertex)

                # get timings

                # check config params for better performance
                (n_at_same_time, local_clocks) = self._auto_config_times(
                    app_machine_quantity, clocks_between_cores,
                    clocks_for_sending, app_vertex, n_keys_map,
                    machine_graph, clocks_waiting)
                n_phases, n_slots, clocks_between_phases = \
                    self._generate_times(
                        machine_graph, app_vertex, n_at_same_time,
                        local_clocks, n_keys_map)

                # store in tracker
                app_vertex.set_other_timings(
                    local_clocks, n_slots, clocks_between_phases,
                    n_phases, clocks_per_cycle)

                # test timings
                fraction_of_sending = self._get_fraction_of_sending(
                    n_phases, clocks_between_phases, clocks_for_sending)
                if fraction_of_sending is not None:
                    max_fraction_of_sending = max(
                        max_fraction_of_sending, fraction_of_sending)

        time_scale_factor_needed = time_scale_factor * max_fraction_of_sending
        if max_fraction_of_sending > 1:
            logger.warning(
                "A time scale factor of {} may be needed to run correctly"
                .format(time_scale_factor_needed))
        elif max_fraction_of_sending < 1:
            logger.info(
                "The time scale factor could be reduced to {}"
                .format(time_scale_factor_needed))

        # get initial offset for each app vertex.
        for app_vertex in application_graph.vertices:
            if isinstance(app_vertex, TDMAAwareApplicationVertex):
                initial_offset = self._generate_initial_offset(
                    app_vertex, app_verts, clocks_initial,
                    clocks_waiting)
                app_vertex.set_initial_offset(initial_offset)

    @staticmethod
    def _auto_config_times(
            app_machine_quantity, clocks_between_cores, clocks_for_sending,
            app_vertex, n_keys_map, machine_graph, clocks_waiting):

        n_cores = app_vertex.get_n_cores()
        n_phases = app_vertex.find_n_phases_for(machine_graph, n_keys_map)

        # If there are no packets sent, pretend there is 1 to avoid division
        # by 0; it won't actually matter anyway
        if n_phases == 0:
            n_phases = 1

        # Overall time of the TDMA window minus initial offset
        overall_clocks_available = clocks_for_sending - clocks_waiting

        # Easier bool compares
        core_set = clocks_between_cores is not None
        app_set = app_machine_quantity is not None

        # Adjust time between cores to fit time scale
        if not core_set and app_set:
            n_slots = int(math.ceil(n_cores / app_machine_quantity))
            clocks_per_phase = (
                int(math.ceil(overall_clocks_available / n_phases)))
            # Note: The plus 1 ensures the last core finishes, if it's the
            # worst in terms of n keys to transmit
            clocks_between_cores = clocks_per_phase / (n_slots + 1)
            logger.debug("adjusted clocks between cores is {}".format(
                clocks_between_cores))

        # Adjust cores at same time to fit time between cores.
        if core_set and not app_set:
            clocks_per_phase = (
                int(math.ceil(overall_clocks_available / n_phases)))
            max_slots = int(math.floor(
                clocks_per_phase / clocks_between_cores))
            app_machine_quantity = int(math.ceil(n_cores / max_slots))
            logger.debug(
                "Adjusted the number of cores of a app vertex that "
                "can fire at the same time to {}".format(
                    app_machine_quantity))

        return app_machine_quantity, clocks_between_cores

    @staticmethod
    def _generate_initial_offset(
            app_vertex, app_verts, clocks_between_cores, clocks_waiting):
        """ Calculates from the app vertex index the initial offset for the\
            TDMA between all cores

        :param ~pacman.model.graphs.application.ApplicationVertex app_vertex:
            the app vertex in question.
        :param app_verts: the list of app vertices.
        :type app_verts:
            list(~pacman.model.graphs.application.ApplicationVertex)
        :param int clocks_between_cores: the clock cycles between cores.
        :param int clocks_waiting: the clock cycles to wait for.
        :return: the initial offset for this app vertex including wait time.
        :rtype: int
        """
        # This is an offset between cores
        initial_offset_clocks = int(math.ceil(
            (app_verts.index(app_vertex) * clocks_between_cores) /
            len(app_verts)))
        # add the offset the clocks to wait for before sending anything at all
        initial_offset_clocks += clocks_waiting
        return initial_offset_clocks

    @staticmethod
    def _generate_times(
            machine_graph, app_vertex, app_machine_quantity,
            clocks_between_cores, n_keys_map):
        """ Generates the number of phases needed for this app vertex, as well\
            as the number of slots and the time between spikes for this app\
            vertex, given the number of machine verts to fire at the same time\
            from a given app vertex.

        :param ~pacman.model.graphs.machine.MachineGraph machine_graph:
            machine graph
        :param TDMAAwareApplicationVertex app_vertex: the app vertex
        :param int app_machine_quantity: the pop spike control level
        :param int clocks_between_cores: the clock cycles between cores
        :param n_keys_map: the partition to n keys map.
        :type n_keys_map:
            ~pacman.model.routing_info.AbstractMachinePartitionNKeysMap
        :return: (n_phases, n_slots, time_between_phases) for this app vertex
        :rtype: tuple(int, int, int)
        """

        # Figure total T2s
        n_phases = app_vertex.find_n_phases_for(machine_graph, n_keys_map)

        # how many hops between T2's
        n_cores = app_vertex.get_n_cores()
        n_slots = int(math.ceil(n_cores / app_machine_quantity))

        # figure T2
        clocks_between_phases = int(math.ceil(clocks_between_cores * n_slots))

        return n_phases, n_slots, clocks_between_phases

    def _get_fraction_of_sending(
            self, n_phases, clocks_between_phases, clocks_for_sending):
        """ Get the fraction of the send

        :param int n_phases:
            the max number of phases this TDMA needs for a given app vertex
        :param int time_between_phases: the time between phases.
        :param float fraction_of_sending:
            fraction of time step for sending packets
        :param str label: the app vertex we're considering at this point
        :return:
        """

        # figure how much time this TDMA needs
        total_clocks_needed = n_phases * clocks_between_phases
        return total_clocks_needed / clocks_for_sending

    def __check_at_most_one(self, name_1, value_1, name_2, value_2):
        if value_1 is not None and value_2 is not None:
            raise ConfigurationException(
                "Both {} and {} have been specified; please choose just one"
                .format(name_1, name_2))

    def __check_only_one(self, name_1, value_1, name_2, value_2):
        """ Checks that exactly one of the values is not None
        """
        self.__check_at_most_one(name_1, value_1, name_2, value_2)
        if value_1 is None and value_2 is None:
            raise ConfigurationException(
                "Exactly one of {} and {} must be specified".format(
                    name_1, name_2))

    def config_values(self, clocks_per_cycle):
        """ Read the config for the right parameters and combinations.

        :param int clocks_per_cycle: The number of clock cycles per time step
        :return: (app_machine_quantity, clocks_between_cores,
                clocks_for_sending, clocks_waiting, initial_clocks)
        :rtype: tuple(int, int, int, int. int)
        """

        # get config
        config = get_simulator().config

        # set the number of cores expected to fire at any given time
        app_machine_quantity = read_config_int(
            config, "Simulation", "app_machine_quantity")

        # set the time between cores to fire
        time_between_cores = read_config_float(
            config, "Simulation", "time_between_cores")
        clocks_between_cores = read_config_int(
            config, "Simulation", "clock_cycles_between_cores")
        self.__check_at_most_one(
            "time_between_cores", time_between_cores,
            "clock_cycles_betwen_cores", clocks_between_cores)
        if (time_between_cores is None and clocks_between_cores is None and
                app_machine_quantity is None):
            raise ConfigurationException(
                "Either one of time_between_cores and clocks_between_cores"
                " must be specified or else app_machine_quantity must be"
                " specified")
        if time_between_cores is not None:
            clocks_between_cores = time_between_cores * CLOCKS_PER_US

        # time spend sending
        fraction_of_sending = read_config_float(
            config, "Simulation", "fraction_of_time_spike_sending")
        clocks_for_sending = read_config_int(
            config, "Simulation", "clock_cycles_sending")
        self.__check_only_one(
            "fraction_of_time_spike_sending", fraction_of_sending,
            "clock_cycles_sending", clocks_for_sending)
        if fraction_of_sending is not None:
            clocks_for_sending = int(round(
                clocks_per_cycle * fraction_of_sending))

        # time waiting before sending
        fraction_of_waiting = read_config_float(
            config, "Simulation", "fraction_of_time_before_sending")
        clocks_waiting = read_config_int(
            config, "Simulation", "clock_cycles_before_sending")
        self.__check_only_one(
            "fraction_of_time_before_sending", fraction_of_waiting,
            "clock_cycles_before_sending", clocks_waiting)
        if fraction_of_waiting is not None:
            clocks_waiting = int(round(clocks_per_cycle * fraction_of_waiting))

        # time to offset app vertices between each other
        fraction_initial = read_config_float(
            config, "Simulation", "fraction_of_time_for_offset")
        clocks_initial = read_config_int(
            config, "Simulation", "clock_cycles_for_offset")
        self.__check_only_one(
            "fraction_of_time_for_offset", fraction_initial,
            "clock_cycles_for_offset", clocks_initial)
        if fraction_initial is not None:
            clocks_initial = int(round(clocks_per_cycle * fraction_initial))

        # check fractions less than 1.
        if (clocks_for_sending + clocks_waiting + clocks_initial >
                clocks_per_cycle):
            raise ConfigurationException(
                "The total time for the TDMA must not exceed the time per"
                " cycle")

        return (app_machine_quantity, clocks_between_cores,
                clocks_for_sending, clocks_waiting, clocks_initial)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import re
from spinn_utilities.ordered_set import OrderedSet
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB

_DDL_FILE = os.path.join(os.path.dirname(__file__), "db.sql")
_RE = re.compile(r"(\d+)([_,:])(\d+)(?:\2(\d+))?")


class SqlLiteDatabase(SQLiteDB):
    """ Specific implementation of the Database for SQLite 3.

    .. note::
        *Not thread safe on the same database file.*
        Threads can access different DBs just fine.

    .. note::
        This totally relies on the way SQLite's type affinities function.
        You can't port to a different database engine without a lot of work.
    """

    __slots__ = []

    def __init__(self, database_file=None):
        """
        :param str database_file:
            The name of a file that contains (or will contain) an SQLite
            database holding the data. If omitted, an unshared in-memory
            database will be used (suitable only for testing).
        :type database_file: str
        """
        super().__init__(database_file, ddl_file=_DDL_FILE)

    def insert_items(self, items):
        """ Does a bulk insert of the items into the database.

        :param list(ProvenanceDataItem) items: The items to insert
        """
        with self.transaction() as cur:
            cur.executemany(
                """
                INSERT OR IGNORE INTO source(
                    source_name, source_short_name, x, y, p)
                VALUES(?, ?, ?, ?, ?)
                """, self.__unique_sources(items, slice(None, -1), "/"))
            cur.executemany(
                """
                INSERT OR IGNORE INTO description(description_name)
                VALUES(?)
                """, self.__unique_names(items, -1))
            cur.executemany(
                """
                INSERT INTO provenance(
                    source_id, description_id, the_value)
                VALUES((
                    SELECT source_id FROM source WHERE source_name = ?
                ), (
                    SELECT description_id FROM description
                    WHERE description_name = ?
                ), ?)
                """, map(self.__condition_row, items))

    @classmethod
    def __unique_names(cls, items, index):
        """ Produces an iterable of 1-tuples of the *unique* names in at \
            particular index into the provenance items' names.

        :param iterable(ProvenanceDataItem) items: The prov items
        :param int index: The index into the names
        :rtype: iterable(tuple(str))
        """
        return ((name,) for name in OrderedSet(
            item.names[index] for item in items))

    @classmethod
    def __unique_sources(cls, items, index, joiner):
        """ Produces an iterable of 1-tuples of the *unique* names in at \
            particular index into the provenance items' names.

        :param iterable(ProvenanceDataItem) items: The prov items
        :param index: The index into the names
        :type index: int or slice
        :param str joiner: Used to make compound names when slices are used
        :rtype: iterable(tuple(str,str,int or None,int or None,int or None))
        """
        if isinstance(index, int):
            return (cls.__coordify(name) for name in OrderedSet(
                item.names[index] for item in items))
        return (cls.__coordify(name, joiner) for name in OrderedSet(
            joiner.join(item.names[index]) for item in items))

    @staticmethod
    def __coordify(name, joiner=None):
        """ Creates the tuple of values for insertion into the source table.

        :param str name: The extracted, possibly compound name
        :rtype: tuple(str, str, int or None, int or None, int or None)
        """
        x = None
        y = None
        p = None
        short_name = name.split(joiner, 1)[0]
        match = _RE.search(name)
        if match:
            x = int(match.group(1))
            y = int(match.group(3))
            if match.group(4):
                p = int(match.group(4))
        return (name, short_name, x, y, p)

    @staticmethod
    def __condition_row(item, joiner="/"):
        """ Converts a provenance item into the form ready for insert.

        .. note::
            This totally relies on the way SQLite's type affinities work.
            In particular, we can store any old item type, which is very
            convenient!

        :param ProvenanceDataItem item: The prov item
        :rtype: tuple(str, str, int or float or str)
        """
        value = item.value
        if isinstance(value, datetime.timedelta):
            value = value.microseconds
        return joiner.join(item.names[:-1]), item.names[-1], value



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class AbstractProvidesLocalProvenanceData(object, metaclass=AbstractBase):
    """ Indicates an object that provides locally obtained provenance data.

    GraphProvenanceGatherer will check all Vertices and all Edges in both the
    MachineGraph and if it exists the ApplicationGraph
    """

    __slots__ = ()

    @abstractmethod
    def get_local_provenance_data(self):
        """ Get an iterable of provenance data items.

        :return: the provenance items
        :rtype:
            iterable(~spinn_front_end_common.utilities.utility_objs.ProvenanceDataItem)
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from spinnman.model.enums import CPUState


class ExecutableType(Enum):
    """ The different types of executable from the perspective of how they\
        are started and controlled.
    """

    #: Runs immediately without waiting for barrier and then exits.
    RUNNING = (
        0,
        [CPUState.RUNNING],
        [CPUState.FINISHED],
        False,
        "Runs immediately without waiting for barrier and then exits")
    #: Calls ``spin1_start(SYNC_WAIT)`` and then eventually ``spin1_exit()``.
    SYNC = (
        1,
        [CPUState.SYNC0],
        [CPUState.FINISHED],
        False,
        "Calls spin1_start(SYNC_WAIT) and then eventually spin1_exit()")
    #: Calls ``simulation_run()`` and ``simulation_exit()`` /
    #: ``simulation_handle_pause_resume()``.
    USES_SIMULATION_INTERFACE = (
        2,
        [CPUState.SYNC0, CPUState.SYNC1, CPUState.PAUSED, CPUState.READY],
        [CPUState.READY],
        True,
        "Calls simulation_run() and simulation_exit() / "
        "simulation_handle_pause_resume()")
    #: Situation where there user has supplied no application but for some
    #: reason still wants to run.
    NO_APPLICATION = (
        3,
        [],
        [],
        True,
        "Situation where there user has supplied no application but for "
        "some reason still wants to run")
    #: Runs immediately without waiting for barrier and never ends.
    SYSTEM = (
        4,
        [CPUState.RUNNING],
        [CPUState.RUNNING],
        True,
        "Runs immediately without waiting for barrier and never ends")

    def __new__(cls, value, start_state, end_state,
                supports_auto_pause_and_resume, doc=""):
        # pylint: disable=protected-access, too-many-arguments
        obj = object.__new__(cls)
        obj._value_ = value
        obj.start_state = start_state
        obj.end_state = end_state
        obj.supports_auto_pause_and_resume = supports_auto_pause_and_resume
        obj.__doc__ = doc
        return obj

    def __init__(self, value, start_state, end_state,
                 supports_auto_pause_and_resume, doc=""):
        # pylint: disable=too-many-arguments
        self._value_ = value
        self.__doc__ = doc
        self.start_state = start_state
        self.end_state = end_state
        self.supports_auto_pause_and_resume = supports_auto_pause_and_resume
        self.__doc__ = doc



# Copyright (c) 2017-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from spinn_utilities.log import FormatAdapter
from spinnman.model import ExecutableTargets, CPUInfos
from spinnman.model.enums import CPUState
from .iobuf_extractor import IOBufExtractor
from .globals_variables import get_simulator

logger = FormatAdapter(logging.getLogger(__name__))


def _emergency_state_check(txrx, app_id):
    """
    :param ~.Transceiver txrx:
    :param int app_id:
    """
    # pylint: disable=broad-except
    try:
        rte_count = txrx.get_core_state_count(
            app_id, CPUState.RUN_TIME_EXCEPTION)
        watchdog_count = txrx.get_core_state_count(app_id, CPUState.WATCHDOG)
        if rte_count or watchdog_count:
            logger.warning(
                "unexpected core states (rte={}, wdog={})",
                txrx.get_cores_in_state(None, CPUState.RUN_TIME_EXCEPTION),
                txrx.get_cores_in_state(None, CPUState.WATCHDOG))
    except Exception:
        logger.exception(
            "Could not read the status count - going to individual cores")
        machine = txrx.get_machine_details()
        infos = CPUInfos()
        errors = list()
        for chip in machine.chips:
            for p in chip.processors:
                try:
                    info = txrx.get_cpu_information_from_core(
                        chip.x, chip.y, p)
                    if info.state in (
                            CPUState.RUN_TIME_EXCEPTION, CPUState.WATCHDOG):
                        infos.add_processor(chip.x, chip.y, p, info)
                except Exception:
                    errors.append((chip.x, chip.y, p))
        logger.warning(txrx.get_core_status_string(infos))
        logger.warning("Could not read information from cores {}".format(
            errors))


# TRICKY POINT: Have to delay the import to here because of import circularity
def _emergency_iobuf_extract(txrx, executable_targets):
    """
    :param ~.Transceiver txrx:
    :param ExecutableTargets executable_targets:
    """
    # pylint: disable=protected-access
    sim = get_simulator()
    extractor = IOBufExtractor(
        txrx, executable_targets, sim._executable_finder,
        sim._app_provenance_file_path, sim._system_provenance_file_path,
        recovery_mode=True, filename_template="emergency_iobuf_{}_{}_{}.txt")
    extractor.extract_iobuf()


def emergency_recover_state_from_failure(txrx, app_id, vertex, placement):
    """ Used to get at least *some* information out of a core when something\
        goes badly wrong. Not a replacement for what abstract spinnaker base\
        does.

    :param ~spinnman.transceiver.Transceiver txrx: The transceiver.
    :param int app_id: The ID of the application.
    :param AbstractHasAssociatedBinary vertex:
        The vertex to retrieve the IOBUF from if it is suspected as being dead
    :param ~pacman.model.placements.Placement placement:
        Where the vertex is located.
    """
    # pylint: disable=protected-access
    _emergency_state_check(txrx, app_id)
    target = ExecutableTargets()
    path = get_simulator()._executable_finder.get_executable_path(
        vertex.get_binary_file_name())
    target.add_processor(
        path, placement.x, placement.y, placement.p,
        vertex.get_binary_start_type())
    _emergency_iobuf_extract(txrx, target)


def emergency_recover_states_from_failure(txrx, app_id, executable_targets):
    """ Used to get at least *some* information out of a core when something\
        goes badly wrong. Not a replacement for what abstract spinnaker base\
        does.

    :param ~spinnman.transceiver.Transceiver txrx: The transceiver.
    :param int app_id: The ID of the application.
    :param ~spinnman.model.ExecutableTargets executable_targets:
        The what/where mapping
    """
    _emergency_state_check(txrx, app_id)
    _emergency_iobuf_extract(txrx, executable_targets)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict, OrderedDict
import json
import logging
import os
import subprocess
from spinn_utilities.log import FormatAdapter
from pacman.exceptions import PacmanExternalAlgorithmFailedToCompleteException
from pacman.operations.algorithm_reports.write_json_machine import (
    WriteJsonMachine)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)

logger = FormatAdapter(logging.getLogger(__name__))


class JavaCaller(object):
    """ Support class that holds all the stuff for running stuff in Java.
        This includes the work of preparing data for transmitting to Java and\
        back.

        This separates the choices of how to call the Java batch vs streaming,\
        jar locations, parameters, etc from the rest of the python code.
    """

    __slots__ = [
        "_chipxy_by_ethernet",
        # The folder holding sqlite databases etc.
        "_report_folder",
        # The call to get java to work. Including the path if required.
        "_java_call",
        # The local https://github.com/SpiNNakerManchester/JavaSpiNNaker
        "_java_spinnaker_path",
        # the folder to write the any json files into
        "_json_folder",
        # The machine
        "_machine",
        # The location where the machine json is written
        "_machine_json_path",
        # Dict of chip (x, y) to the p of the monitor vertex
        "_monitor_cores",
        # Flag to indicate if at least one placement is recording
        "_recording",
        # Dict of ethernet (x, y) and the packetGather IPtago
        "_gatherer_iptags",
        # Dict of ethernet (x, y) to the p of the packetGather vertex
        "_gatherer_cores",
        # The location where the latest placement json is written
        "_placement_json",
        # Properties flag to be passed to Java
        "_java_properties"
    ]

    def __init__(self, json_folder, java_call, java_spinnaker_path=None,
                 java_properties=None):
        """ Creates a java caller and checks the user/config parameters.

        :param str json_folder: The location where the machine JSON is written.
        :param str java_call:
            Call to start java. Including the path if required.
        :param str java_spinnaker_path:
            The path where the java code can be found.
            This must point to a local copy of
            `https://github.com/SpiNNakerManchester/JavaSpiNNaker`.
            It must also have been built!
            If `None` the assumption is that it is the same parent directory as
            `https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon`.
        :param str java_properties:
            Optional properties that will be passed to Java.\
            Must start with ``-D``.
            For example ``-Dlogging.level=DEBUG``
        :raise ConfigurationException: if simple parameter checking fails.
        """
        self._recording = None
        self._report_folder = None
        self._json_folder = json_folder

        self._java_call = java_call
        result = subprocess.call([self._java_call, '-version'])
        if result != 0:
            raise ConfigurationException(
                " {} -version failed. "
                "Please set [Java] java_call to the absolute path "
                "to start java. (in config file)".format(self._java_call))

        if java_spinnaker_path is None:
            interface = os.path.dirname(os.path.realpath(__file__))
            spinn_front_end_common = os.path.dirname(interface)
            github_checkout_dir = os.path.dirname(spinn_front_end_common)
            parent = os.path.dirname(github_checkout_dir)
            self._java_spinnaker_path = os.path.join(
                parent, "JavaSpiNNaker")
        else:
            # As I don't know how to write pwd and /JavaSpiNNaker to one line
            indirect_path = os.path.join(
                java_spinnaker_path, "JavaSpiNNaker")
            if os.path.isdir(indirect_path):
                self._java_spinnaker_path = indirect_path
            else:
                self._java_spinnaker_path = java_spinnaker_path
        if not os.path.isdir(self._java_spinnaker_path):
            raise ConfigurationException(
                "No Java code found at {}".format(self._java_spinnaker_path))

        self._machine = None
        self._machine_json_path = None
        self._placement_json = None
        self._monitor_cores = None
        self._gatherer_iptags = None
        self._gatherer_cores = None
        self._java_properties = java_properties
        self._chipxy_by_ethernet = None
        if self._java_properties is not None:
            self._java_properties = self._java_properties.split()
            for _property in self._java_properties:
                if _property[:2] != "-D":
                    raise ConfigurationException(
                        "Java Properties must start with -D found at {}".
                        format(_property))

    def set_machine(self, machine):
        """ Passes the machine in leaving this class to decide pass it to Java.

        :param ~spinn_machine.Machine machine: A machine Object
        """
        self._machine = machine

    def set_advanced_monitors(
            self, placements, tags, monitor_cores, packet_gathers):
        """
        :param ~pacman.model.placements.Placements placements:
            The placements of the vertices
        :param ~pacman.model.tags.Tags tags: The tags assigned to the vertices
        :param monitor_cores: Where the advanced monitor for each core is
        :type monitor_cores:
            dict(tuple(int,int), ExtraMonitorSupportMachineVertex)
        :param packet_gathers: Where the packet gatherers are
        :type packet_gathers:
            dict(tuple(int,int), DataSpeedUpPacketGatherMachineVertex)
        :rtype: None
        """
        self._monitor_cores = dict()
        for core, monitor_core in monitor_cores.items():
            placement = placements.get_placement_of_vertex(monitor_core)
            self._monitor_cores[core] = placement.p

        self._gatherer_iptags = dict()
        self._gatherer_cores = dict()
        for core, packet_gather in packet_gathers.items():
            self._gatherer_iptags[core] = \
                tags.get_ip_tags_for_vertex(packet_gather)[0]
            placement = placements.get_placement_of_vertex(packet_gather)
            self._gatherer_cores[core] = placement.p

        self._chipxy_by_ethernet = defaultdict(list)
        for chip in self._machine.chips:
            if not chip.virtual:
                chip_xy = (chip.x, chip.y)
                ethernet = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
                self._chipxy_by_ethernet[ethernet].append(chip_xy)

    def _machine_json(self):
        """ Converts the machine in this class to JSON.

        :return: the name of the file containing the JSON
        """
        if self._machine_json_path is None:
            self._machine_json_path = WriteJsonMachine.write_json(
                self._machine, self._json_folder)
        return self._machine_json_path

    def set_report_folder(self, report_folder):
        """ Passes the database file in.

        :param str report_folder:
            Path to directory with SQLite databases and into which java will
            write.
        """
        self._report_folder = report_folder

    def set_placements(self, placements, transceiver):
        """ Passes in the placements leaving this class to decide pass it to
            Java.

        This method may obtain extra information about he placements which is
        why it also needs the transceiver.

        Currently the extra information extracted is recording region base
        address but this could change if recording region saved in the
        database.

        Currently this method uses JSON but that may well change to using the
        database.

        :param ~pacman.model.placements.Placements placements:
            The Placements Object
        :param ~spinnman.transceiver.Transceiver transceiver: The Transceiver
        """
        path = os.path.join(self._json_folder, "java_placements.json")
        self._recording = False
        if self._gatherer_iptags is None:
            self._placement_json = self._write_placements(
                placements, transceiver, path)
        else:
            self._placement_json = self._write_gather(
                placements, transceiver, path)

    def _json_placement(self, placement, transceiver):
        """
        :param ~pacman.model.placements.Placement placement:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :rtype: dict
        """
        vertex = placement.vertex
        json_placement = OrderedDict()
        json_placement["x"] = placement.x
        json_placement["y"] = placement.y
        json_placement["p"] = placement.p

        json_vertex = OrderedDict()
        json_vertex["label"] = vertex.label
        if isinstance(vertex, AbstractReceiveBuffersToHost) and \
                vertex.get_recorded_region_ids():
            self._recording = True
            json_vertex["recordedRegionIds"] = vertex.get_recorded_region_ids()
            json_vertex["recordingRegionBaseAddress"] = \
                vertex.get_recording_region_base_address(
                    transceiver, placement)
        else:
            json_vertex["recordedRegionIds"] = []
            json_vertex["recordingRegionBaseAddress"] = 0
        json_placement["vertex"] = json_vertex

        return json_placement

    def _json_iptag(self, iptag):
        """
        :param ~pacman.model.tags.IPTag iptag:
        :rtype: dict
        """
        json_tag = OrderedDict()
        json_tag["x"] = iptag.destination_x
        json_tag["y"] = iptag.destination_y
        json_tag["boardAddress"] = iptag.board_address
        json_tag["targetAddress"] = iptag.ip_address
        # Intentionally not including port!
        json_tag["stripSDP"] = iptag.strip_sdp
        json_tag["tagID"] = iptag.tag
        json_tag["trafficIdentifier"] = iptag.traffic_identifier

        return json_tag

    def _placements_grouped(self, placements):
        """
        :param ~pacman.model.placements.Placements placements:
        :rtype: dict(tuple(int,int),dict(tuple(int,int),
            ~pacman.model.placements.Placement))
        """
        by_ethernet = defaultdict(lambda: defaultdict(list))
        for placement in placements:
            chip = self._machine.get_chip_at(placement.x, placement.y)
            chip_xy = (placement.x, placement.y)
            ethernet = (chip.nearest_ethernet_x, chip.nearest_ethernet_y)
            by_ethernet[ethernet][chip_xy].append(placement)
        return by_ethernet

    def _write_gather(self, placements, transceiver, path):
        """
        :param ~pacman.model.placements.Placements placements:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param str path:
        :rtype: str
        """
        placements_by_ethernet = self._placements_grouped(placements)
        json_obj = list()
        for ethernet in self._chipxy_by_ethernet:
            by_chip = placements_by_ethernet[ethernet]
            json_gather = OrderedDict()
            json_gather["x"] = ethernet[0]
            json_gather["y"] = ethernet[1]
            json_gather["p"] = self._gatherer_cores[ethernet]
            json_gather["iptag"] = self._json_iptag(
                self._gatherer_iptags[ethernet])
            json_chips = list()
            for chip_xy in self._chipxy_by_ethernet[ethernet]:
                json_chip = OrderedDict()
                json_chip["x"] = chip_xy[0]
                json_chip["y"] = chip_xy[1]
                json_chip["p"] = self._monitor_cores[chip_xy]
                if chip_xy in by_chip:
                    json_placements = list()
                    for placement in by_chip[chip_xy]:
                        json_p = self._json_placement(placement, transceiver)
                        if json_p:
                            json_placements.append(json_p)
                    if len(json_placements) > 0:
                        json_chip["placements"] = json_placements
                json_chips.append(json_chip)
            json_gather["monitors"] = json_chips
            json_obj.append(json_gather)

        # dump to json file
        with open(path, "w") as f:
            json.dump(json_obj, f)

        return path

    def _write_placements(self, placements, transceiver, path):
        """
        :param ~pacman.model.placements.Placements placements:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param str path:
        :rtype: str
        """
        # Read back the regions
        json_obj = list()
        for placement in placements:
            json_p = self._json_placement(placement, transceiver)
            if json_p:
                json_obj.append(json_p)

        # dump to json file
        with open(path, "w") as f:
            json.dump(json_obj, f)

        return path

    @property
    def _jar_file(self):
        """ The fully qualified name of the JavaSpiNNaker executable JAR.

        :rtype: str
        """
        f = os.path.join(
            self._java_spinnaker_path, "SpiNNaker-front-end",
            "target", "spinnaker-exe.jar")
        if not os.path.exists(f):
            logger.warning("no Java build in file {}; failure expected", f)
        return f

    def _run_java(self, *args):
        """ Does the actual running of JavaSpiNNaker. Arguments are those that
            will be processed by the `main` method on the Java side.

        :rtype: int
        """
        if self._java_properties is None:
            params = [self._java_call, '-jar', self._jar_file]
        else:
            params = [self._java_call] + self._java_properties \
                     + ['-jar', self._jar_file]
        params.extend(args)
        return subprocess.call(params)

    def get_all_data(self):
        """ Gets all the data from the previously set placements
            and put these in the previously set database.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        if not self._recording:
            return

        if self._gatherer_iptags is None:
            result = self._run_java(
                'download', self._placement_json, self._machine_json(),
                self._report_folder)
        else:
            result = self._run_java(
                'gather', self._placement_json, self._machine_json(),
                self._report_folder)
        if result != 0:
            log_file = os.path.join(self._report_folder, "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_data_specification(self):
        """ Writes all the data specs, uploading the result to the machine.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        result = self._run_java(
            'dse', self._machine_json(), self._report_folder)
        if result != 0:
            log_file = os.path.join(self._report_folder, "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_system_data_specification(self):
        """ Writes all the data specs for system cores,
            uploading the result to the machine.

        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        result = self._run_java(
            'dse_sys', self._machine_json(), self._report_folder)
        if result != 0:
            log_file = os.path.join(self._report_folder, "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")

    def execute_app_data_specification(self, use_monitors):
        """ Writes all the data specs for application cores,
            uploading the result to the machine.

        .. note:
            May assume that system cores are already loaded and running if
            `use_monitors` is set to `True`.

        :param bool use_monitors:
        :raises PacmanExternalAlgorithmFailedToCompleteException:
            On failure of the Java code.
        """
        if use_monitors:
            result = self._run_java(
                'dse_app_mon', self._placement_json, self._machine_json(),
                self._report_folder, self._report_folder)
        else:
            result = self._run_java(
                'dse_app', self._machine_json(), self._report_folder)
        if result != 0:
            log_file = os.path.join(self._report_folder, "jspin.log")
            raise PacmanExternalAlgorithmFailedToCompleteException(
                "Java call exited with value " + str(result) + " see "
                + str(log_file) + " for logged info")



# Copyright (c) 2017-2020 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re
from spinn_utilities.log import FormatAdapter
from spinn_utilities.make_tools.replacer import Replacer
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine.core_subsets import CoreSubsets
from spinnman.model.io_buffer import IOBuffer
from spinn_front_end_common.utilities.utility_objs import ExecutableType
from spinn_front_end_common.utilities.helpful_functions import (
    convert_string_into_chip_and_core_subset)

logger = FormatAdapter(logging.getLogger(__name__))
ERROR_ENTRY = re.compile(r"\[ERROR\]\s+\((.*)\):\s+(.*)")
WARNING_ENTRY = re.compile(r"\[WARNING\]\s+\((.*)\):\s+(.*)")
ENTRY_FILE = 1
ENTRY_TEXT = 2


class _DummyProgress(object):
    def over(self, values):
        return values


class IOBufExtractor(object):
    """ Extract the logging output buffers from the machine, and separates\
        lines based on their prefix.
    """

    __slots__ = ["_filename_template", "_recovery_mode", "__system_binaries",
                 "__app_path", "__sys_path", "__transceiver",
                 "__suppress_progress", "__from_cores", "__binary_types",
                 "__executable_targets", "__executable_finder"]

    def __init__(self, transceiver, executable_targets,
                 executable_finder, app_provenance_file_path,
                 system_provenance_file_path, from_cores="ALL",
                 binary_types=None, recovery_mode=False,
                 filename_template="iobuf_for_chip_{}_{}_processor_id_{}.txt",
                 suppress_progress=False):
        """
        :param bool recovery_mode:
        :param str filename_template:
        :param bool suppress_progress:
        :param ~spinnman.transceiver.Transceiver transceiver:
        :param ~spinnman.model.ExecutableTargets executable_targets:
        :param ExecutableFinder executable_finder:
        :param app_provenance_file_path:
        :type app_provenance_file_path: str or None
        :param system_provenance_file_path:
        :type system_provenance_file_path: str or None
        :param str from_cores:
        :param str binary_types:
        """
        self._filename_template = filename_template
        self._recovery_mode = bool(recovery_mode)
        self.__suppress_progress = bool(suppress_progress)

        self.__app_path = app_provenance_file_path
        self.__sys_path = system_provenance_file_path
        self.__transceiver = transceiver
        self.__from_cores = from_cores
        self.__binary_types = binary_types
        self.__executable_targets = executable_targets
        self.__executable_finder = executable_finder

        self.__system_binaries = set()
        try:
            self.__system_binaries.update(
                executable_targets.get_binaries_of_executable_type(
                    ExecutableType.SYSTEM))
        except KeyError:
            pass

    def extract_iobuf(self):
        """ Perform the extraction of IOBUF

        :return: error_entries, warn_entries
        :rtype: tuple(list(str),list(str))
        """
        if self.__from_cores == "ALL":
            return self.__extract_all_cores()
        elif self.__from_cores and self.__binary_types:
            return self.__extract_selected_cores_and_types()
        elif self.__from_cores:
            return self.__extract_selected_cores()
        elif self.__binary_types:
            return self.__extract_selected_types()
        else:
            # nothing
            return [], []

    def __progress(self, bins):
        """
        :param list bins:
        :rtype: ~.ProgressBar
        """
        if self.__suppress_progress:
            return _DummyProgress()
        label = (("Recovering" if self._recovery_mode else "Extracting")
                 + " IOBUF from the machine")
        return ProgressBar(len(bins), label)

    def __prov_path(self, binary):
        """
        :param str binary:
        :return: provenance directory path
        :rtype: str
        """
        return (self.__sys_path if binary in self.__system_binaries
                else self.__app_path)

    def __extract_all_cores(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # all the cores
        progress = self.__progress(self.__executable_targets.binaries)
        for binary in progress.over(self.__executable_targets.binaries):
            core_subsets = self.__executable_targets.get_cores_for_binary(
                binary)
            self.__extract_iobufs_for_binary(
                core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_cores_and_types(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # bit of both
        progress = self.__progress(self.__executable_targets.binaries)
        binaries = self.__executable_finder.get_executable_paths(
            self.__binary_types)
        iocores = convert_string_into_chip_and_core_subset(self.__from_cores)
        for binary in progress.over(self.__executable_targets.binaries):
            if binary in binaries:
                core_subsets = self.__executable_targets.get_cores_for_binary(
                    binary)
            else:
                core_subsets = iocores.intersect(
                    self.__executable_targets.get_cores_for_binary(binary))
            if core_subsets:
                self.__extract_iobufs_for_binary(
                    core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_cores(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # some hard coded cores
        progress = self.__progress(self.__executable_targets.binaries)
        iocores = convert_string_into_chip_and_core_subset(self.__from_cores)
        for binary in progress.over(self.__executable_targets.binaries):
            core_subsets = iocores.intersect(
                self.__executable_targets.get_cores_for_binary(binary))
            if core_subsets:
                self.__extract_iobufs_for_binary(
                    core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_selected_types(self):
        """
        :rtype: tuple(list(str), list(str))
        """
        error_entries = list()
        warn_entries = list()
        # some binaries
        binaries = self.__executable_finder.get_executable_paths(
            self.__binary_types)
        progress = self.__progress(binaries)
        for binary in progress.over(binaries):
            core_subsets = self.__executable_targets.get_cores_for_binary(
                binary)
            self.__extract_iobufs_for_binary(
                core_subsets, binary, error_entries, warn_entries)
        return error_entries, warn_entries

    def __extract_iobufs_for_binary(
            self, core_subsets, binary, error_entries, warn_entries):
        """
        :param ~.CoreSubsets core_subsets: Where the binary is deployed
        :param str binary: What binary was deployed there.
            This is used to determine how to decompress the IOBUF output.
        :param list(str) error_entries:
        :param list(str) warn_entries:
        """
        replacer = Replacer(binary)
        prov_path = self.__prov_path(binary)

        # extract iobuf
        if self._recovery_mode:
            io_buffers = self.__recover_iobufs(core_subsets)
        else:
            io_buffers = list(self.__transceiver.get_iobuf(core_subsets))

        # write iobuf
        for iobuf in io_buffers:
            self.__process_one_iobuf(iobuf, prov_path, replacer,
                                     error_entries, warn_entries)

    def __process_one_iobuf(
            self, iobuf, file_path, replacer, error_entries, warn_entries):
        """
        :param ~.IOBuffer iobuf:
        :param str file_path:
        :param ~.Replacer replacer:
        :param list(str) error_entries:
        :param list(str) warn_entries:
        """
        file_name = os.path.join(
            file_path, self._filename_template.format(
                iobuf.x, iobuf.y, iobuf.p))

        # set mode of the file based off if the file already exists
        mode = "a" if os.path.exists(file_name) else "w"

        # write iobuf to file and call out errors and warnings.
        with open(file_name, mode) as f:
            for line in iobuf.iobuf.split("\n"):
                replaced = replacer.replace(line)
                f.write(replaced)
                f.write("\n")
                self.__add_value_if_match(
                    ERROR_ENTRY, replaced, error_entries, iobuf)
                self.__add_value_if_match(
                    WARNING_ENTRY, replaced, warn_entries, iobuf)

    def __recover_iobufs(self, core_subsets):
        """
        :param ~.CoreSubsets core_subsets:
        :rtype: list(~.IOBuffer)
        """
        io_buffers = []
        for core_subset in core_subsets:
            for p in core_subset.processor_ids:
                cs = CoreSubsets()
                cs.add_processor(core_subset.x, core_subset.y, p)
                try:
                    io_buffers.extend(self.__transceiver.get_iobuf(cs))
                except Exception as e:  # pylint: disable=broad-except
                    io_buffers.append(IOBuffer(
                        core_subset.x, core_subset.y, p,
                        "failed to retrieve iobufs from {},{},{}; {}".format(
                            core_subset.x, core_subset.y, p, str(e))))
        return io_buffers

    @staticmethod
    def __add_value_if_match(regex, line, entries, iobuf):
        """
        :param ~typing.Pattern regex:
        :param str line:
        :param list(str) entries:
        :param ~.IOBuffer iobuf:
        """
        match = regex.match(line)
        if match:
            entries.append("{}, {}, {}: {} ({})".format(
                iobuf.x, iobuf.y, iobuf.p, match.group(ENTRY_TEXT),
                match.group(ENTRY_FILE)))



# Copyright (c) 2020-2021 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from spinn_utilities.log import FormatAdapter
from pacman.model.partitioner_splitters import (
    SplitterOneAppOneMachine, SplitterOneToOneLegacy, SplitterSliceLegacy)
from pacman.model.graphs.application.abstract import (
    AbstractOneAppOneMachineVertex)
from spinn_front_end_common.utility_models import (
    ReverseIpTagMultiCastSource, LivePacketGather, ChipPowerMonitor)

logger = FormatAdapter(logging.getLogger(__name__))


class SplitterSelector(object):
    """ Splitter object selector that allocates nothing but legacy\
        splitter objects where required
    """

    NOT_KNOWN_APP_VERTEX_ERROR_MESSAGE = (
        "The SplitterSelector has not seen the {} vertex before. Therefore "
        "there is no known splitter to allocate to this app vertex and so "
        "will use the SplitterSliceLegacy Splitter.")

    def __call__(self, app_graph):
        """ basic selector which puts the legacy splitter object on\
            everything without a splitter object

        :param ApplicationGraph app_graph: app graph
        :rtype: None
        """
        for app_vertex in app_graph.vertices:
            if app_vertex.splitter is None:
                self.vertex_selector(app_vertex)

    def vertex_selector(self, app_vertex):
        """ main point for selecting a splitter object for a given app vertex.

        Will assume the SplitterSliceLegacy if no heuristic is known for the
        app vertex.

        :param ~pacman.model.graphs.application.ApplicationVertex app_vertex:
            app vertex to give a splitter object to
        :rtype: None
        """
        if isinstance(app_vertex, AbstractOneAppOneMachineVertex):
            app_vertex.splitter = SplitterOneAppOneMachine()
        elif isinstance(app_vertex, ReverseIpTagMultiCastSource):
            app_vertex.splitter = SplitterSliceLegacy()
        elif isinstance(app_vertex, LivePacketGather):
            app_vertex.splitter = SplitterOneToOneLegacy()
        elif isinstance(app_vertex, ChipPowerMonitor):
            app_vertex.splitter = SplitterOneToOneLegacy()
        else:
            logger.warning(
                self.NOT_KNOWN_APP_VERTEX_ERROR_MESSAGE.format(app_vertex))
            app_vertex.splitter = SplitterSliceLegacy()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSendsBuffersFromHost(object, metaclass=AbstractBase):
    """ Interface to an object that sends buffers of keys to be\
        transmitted at given timestamps in the simulation.
    """

    __slots__ = ()

    @abstractmethod
    def buffering_input(self):
        """ Return True if the input of this vertex is to be buffered.

        :rtype: bool
        """

    @abstractmethod
    def get_regions(self):
        """ Get the set of regions for which there are keys to be sent

        :return: Iterable of region IDs
        :rtype: iterable(int)
        """

    @abstractmethod
    def get_region_buffer_size(self, region):
        """ Get the size of the buffer to be used in SDRAM on the machine\
            for the region in bytes

        :param int region: The region to get the buffer size of
        :return: The size of the buffer space in bytes
        :rtype: int
        """

    @abstractmethod
    def is_next_timestamp(self, region):
        """ Determine if there is another timestamp with data to be sent

        :param int region: The region to determine if there is more data for
        :return: Whether there is more data
        :rtype: bool
        """

    @abstractmethod
    def get_next_timestamp(self, region):
        """ Get the next timestamp at which there are still keys to be sent\
            for the given region

        :param int region: The region to get the timestamp for
        :return: The timestamp of the next available keys
        :rtype: int
        """

    @abstractmethod
    def is_next_key(self, region, timestamp):
        """ Determine if there are still keys to be sent at the given\
            timestamp for the given region

        :param int region: The region to determine if there are keys for
        :param int timestamp:
            The timestamp to determine if there are more keys for
        :return: Whether there are more keys to send for the parameters
        :rtype: bool
        """

    @abstractmethod
    def get_next_key(self, region):
        """ Get the next key in the given region

        :param int region: The region to get the next key from
        :return: The next key, or None if there are no more keys
        :rtype: int
        """

    @abstractmethod
    def is_empty(self, region):
        """ Return true if there are no spikes to be buffered for the\
            specified region

        :param int region: The region to get the next key from
        :return: Whether there are no keys to send for the region
        :rtype: bool
        """

    @abstractmethod
    def rewind(self, region):
        """ Rewinds the internal buffer in preparation of re-sending\
            the spikes

        :param int region: The region to rewind
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .clear_reinjection_queue_message import ClearReinjectionQueueMessage
from .get_reinjection_status_message import (
    GetReinjectionStatusMessage, GetReinjectionStatusMessageResponse)
from .load_application_mc_routes_message import LoadApplicationMCRoutesMessage
from .load_system_mc_routes_message import LoadSystemMCRoutesMessage
from .reset_counters_message import ResetCountersMessage
from .set_reinjection_packet_types_message import (
    SetReinjectionPacketTypesMessage)
from .set_router_timeout_message import SetRouterTimeoutMessage

__all__ = [
    "ClearReinjectionQueueMessage",
    "GetReinjectionStatusMessage",
    "GetReinjectionStatusMessageResponse",
    "LoadApplicationMCRoutesMessage",
    "LoadSystemMCRoutesMessage",
    "ResetCountersMessage",
    "SetReinjectionPacketTypesMessage",
    "SetRouterTimeoutMessage",
    ]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_front_end_common.utilities.constants import (
    SARK_PER_MALLOC_SDRAM_USAGE, BYTES_PER_WORD)

# Size of data in the header for each recording region
# 1 word for space, 1 word for size+missing, 1 word for pointer
_PER_REGION_HEADER_SIZE = BYTES_PER_WORD * 3


def get_recording_header_size(n_recording_regions):
    """ Get the size of the data to be written for the recording header

    :param int n_recording_regions: The number of regions that can be recorded
    :rtype: int
    """
    # See recording.h/recording_initialise for data included in the header
    return BYTES_PER_WORD + (n_recording_regions * _PER_REGION_HEADER_SIZE)


def get_recording_data_constant_size(n_recording_regions):
    """ Get the size of the recorded data to be reserved that doesn't change

    :param int n_recording_regions: The number of regions that can be recorded
    :rtype: int
    """
    return (n_recording_regions * SARK_PER_MALLOC_SDRAM_USAGE)


def get_recording_header_array(recorded_region_sizes):
    """ Get data to be written for the recording header

    :param list(int) recorded_region_sizes:
        A list of sizes of each region to be recorded.
        A size of 0 is acceptable.
    :rtype: list(int)
    """

    # See recording.h/recording_initialise for data included in the header
    data = list()

    # The parameters
    data.append(len(recorded_region_sizes))

    # Add (space, size, pointer) for each region
    for space in recorded_region_sizes:
        data.extend([space, 0, 0])

    return data



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from concurrent.futures import ThreadPoolExecutor, wait
from spinn_utilities.abstract_context_manager import AbstractContextManager
from spinn_utilities.log import FormatAdapter
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.command_messages import (
    DatabaseConfirmation, NotificationProtocolPauseStop,
    NotificationProtocolStartResume)
from spinn_front_end_common.utilities.constants import (
    MAX_DATABASE_PATH_LENGTH)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))


class NotificationProtocol(AbstractContextManager):
    """ The protocol which hand shakes with external devices about the\
        database and starting execution
    """
    __slots__ = [
        "__database_message_connections",
        "__sent_visualisation_confirmation",
        "__socket_addresses",
        "__wait_for_read_confirmation",
        "__wait_futures",
        "__wait_pool"]

    def __init__(self, socket_addresses, wait_for_read_confirmation):
        """
        :param socket_addresses: Where to notify.
        :type socket_addresses:
            list(~spinn_utilities.socket_address.SocketAddress)
        :param bool wait_for_read_confirmation:
            Whether to wait for the other side to acknowledge
        """
        self.__socket_addresses = socket_addresses

        # Determines whether to wait for confirmation that the database
        # has been read before starting the simulation
        self.__wait_for_read_confirmation = wait_for_read_confirmation
        self.__wait_pool = ThreadPoolExecutor(max_workers=1)
        self.__wait_futures = list()
        self.__sent_visualisation_confirmation = False
        self.__database_message_connections = list()
        for socket_address in socket_addresses:
            self.__database_message_connections.append(EIEIOConnection(
                local_port=socket_address.listen_port,
                remote_host=socket_address.notify_host_name,
                remote_port=socket_address.notify_port_no))

    def wait_for_confirmation(self):
        """ If asked to wait for confirmation, waits for all external systems\
            to confirm that they are configured and have read the database

        :rtype: None
        """
        if self.__wait_for_read_confirmation:
            logger.info("** Awaiting for a response from an external source "
                        "to state its ready for the simulation to start **")
            wait(self.__wait_futures)
        self.__wait_futures = list()

    def send_start_resume_notification(self):
        """ Either waits till all sources have confirmed read the database\
            and are configured, and/or just sends the start notification\
            (when the system is executing)

        :rtype: None
        """
        logger.info("** Sending start / resume message to external sources "
                    "to state the simulation has started or resumed. **")
        if self.__wait_for_read_confirmation:
            self.wait_for_confirmation()
        eieio_command_message = NotificationProtocolStartResume()
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(eieio_command_message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to send start/resume notification to external "
                    "application on {}:{} about the simulation ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

    def send_stop_pause_notification(self):
        """ Sends the pause / stop notifications when the script has either\
            finished or paused

        :rtype: None
        """
        logger.info("** Sending pause / stop message to external sources "
                    "to state the simulation has been paused or stopped. **")
        eieio_command_message = NotificationProtocolPauseStop()
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(eieio_command_message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to send stop/pause notification to external "
                    "application on {}:{} about the simulation ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

    # noinspection PyPep8
    def send_read_notification(self, database_path):
        """ Sends notifications to all devices which have expressed an\
            interest in when the database has been written

        :param str database_path: the path to the database file
        """
        notification_thread = self.__wait_pool.submit(
                self._send_read_notification, database_path)
        if self.__wait_for_read_confirmation:
            self.__wait_futures.append(notification_thread)

    def _send_read_notification(self, database_path):
        """ Sends notifications to a list of socket addresses that the\
            database has been written. Message also includes the path to the\
            database

        :param str database_path: the path to the database
        """
        # noinspection PyBroadException
        try:
            self.__do_read_notify(database_path)
        except Exception:  # pylint: disable=broad-except
            logger.warning("problem when sending DB notification",
                           exc_info=True)

    def __do_read_notify(self, database_path):
        # add file path to database into command message.
        if (database_path is not None and
                len(database_path) > MAX_DATABASE_PATH_LENGTH):
            raise ConfigurationException(
                "The file path to the database is too large to be transmitted "
                "via the command packet, please set the file path manually "
                "and set the .cfg parameter [Database] send_file_path to "
                "False")
        message = DatabaseConfirmation(database_path)

        # Send command and wait for response
        logger.info(
            "** Notifying external sources that the database is ready for "
            "reading **")

        # noinspection PyBroadException
        for c in self.__database_message_connections:
            try:
                c.send_eieio_message(message)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to notify external application on {}:{} "
                    "about the database ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

        self.__sent_visualisation_confirmation = True

        # if the system needs to wait, try receiving a packet back
        for c in self.__database_message_connections:
            try:
                if self.__wait_for_read_confirmation:
                    c.receive_eieio_message()
                    logger.info(
                        "** Confirmation from {}:{} received, continuing **",
                        c.remote_ip_address, c.remote_port)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    "*** Failed to receive notification from external "
                    "application on {}:{} about the database ***",
                    c.remote_ip_address, c.remote_port, exc_info=True)

    @property
    def sent_visualisation_confirmation(self):
        """ Whether the external application has actually been notified yet.

        :rtype: bool
        """
        return self.__sent_visualisation_confirmation

    def close(self):
        """ Closes the thread pool
        """
        if self.__wait_pool:
            self.__wait_pool.shutdown()
            self.__wait_futures = list()
            self.__wait_pool = None



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import string
from lxml import etree
from spinn_front_end_common.utilities.helpful_functions import (
    generate_unique_folder_name)

_VALID_CHARS = frozenset(
    "-_.() {}{}".format(string.ascii_letters, string.digits))
_XML_BRANCH_NAME = "provenance_data_items"
_XML_LEAF_NAME = "provenance_data_item"


class ProvenanceXMLWriter(object):
    """ Writes provenance in XML format.
    """

    __slots__ = []

    def __call__(self, provenance_data_items, provenance_data_path):
        """
        :param list(ProvenanceDataItem) provenance_data_items:
            data items for provenance
        :param str provenance_data_path: the file path to store provenance in
        """
        # Group data by the first name
        items = sorted(provenance_data_items, key=lambda item: item.names[0])
        for name, group in itertools.groupby(
                items, lambda item: item.names[0]):
            # Create a root node
            root = etree.Element(_XML_BRANCH_NAME, name=name)
            # Keep track of sub-categories
            categories = {root: dict()}

            # Go through the items and add them
            for item in group:
                # Add the "categories" for the item (any name between the first
                # and last)
                parent = self._build_path(root, categories, item)

                # Add the item
                element = etree.SubElement(
                    parent, _XML_LEAF_NAME, name=item.names[-1])
                element.text = str(item.value)

            # write xml form into file provided
            with open(self._get_file(provenance_data_path, name), "wb") as f:
                f.write(etree.tostring(root, pretty_print=True))

    @staticmethod
    def _get_file(path, name):
        """
        :param str path:
        :param str name:
        :rtype: str
        """
        remapped = "".join(c if c in _VALID_CHARS else '_' for c in name)
        return generate_unique_folder_name(path, remapped, ".xml")

    @staticmethod
    def _build_path(root, categories, item):
        """
        :param etree.Element root:
        :param dict(etree.Element,dict(str,etree.Element)) categories:
        :param ProvenanceDataItem item:
        :rtype: etree.Element
        """
        parent = root
        cats = categories[root]
        for cat_name in item.names[1:-1]:
            if cat_name not in cats:
                # If there is not a category of this name under the parent
                # element, create a new category
                parent = etree.SubElement(
                    parent, _XML_BRANCH_NAME, name=cat_name)
                cats[cat_name] = parent
                categories[parent] = dict()
            # Update our state variables for the next run of the loop
            parent = cats[cat_name]
            cats = categories[parent]
        return parent
